"""lsfHelper.py shouldn't need a batch system and so the unit tests here should aim to run on any system."""
from toil.batchSystems.lsfHelper import parse_mem_and_cmd_from_output
from toil.test import ToilTest


class LSFHelperTest(ToilTest):
    def test_parse_mem_and_cmd_from_output(self):
        # https://github.com/DataBiosphere/toil/pull/3475
        output = ('\nJob <2924748>, Job Name <toil_job_64>, User <thiagogenez>, Project <default>, S'
                  '\n                     tatus <RUN>, Queue <research-rh74>, Job Priority <50>, Com'
                  '\n                     mand <_toil_worker CactusBarRecursion file:/hps/nobackup/p'
                  '\n                     roduction/ensembl/thiagogenez/pairwises/arabidopsis/run/jo'
                  '\n                     bstore/3 kind-CactusBarRecursion/instance-iu6wo56x --conte'
                  '\n                     xt gAShortenedh32xqlE51Yi4=>, Share group charged </thiago'
                  '\n                     genez>, Esub <esub>'
                  '\nThu Mar 18 02:06:32: Submitted from host <noah-login-01>, CWD </hps/nobackup/pr'
                  '\n                     oduction/ensembl/thiagogenez/pairwises/arabidopsis/run>, S'
                  '\n                     pecified CWD </hps/nobackup/production/ensembl/thiagogenez'
                  '\n                     /pairwises/arabidopsis/run/.>, Output File </tmp/toil_work'
                  '\n                     flow_10e83102-2e4b-4093-9128-2a52f4bbc65a_job_64_batch_lsf'
                  '\n                     _2924748_std_output.log>, Error File </tmp/toil_workflow_1'
                  '\n                     0e83102-2e4b-4093-9128-2a52f4bbc65a_job_64_batch_lsf_29247'
                  '\n                     48_std_error.log>, Requested Resources <select[(mem>4000)]'
                  '\n                      rusage[mem=4000:duration=480, numcpus=1:duration=480]>;'
                  '\nThu Mar 18 02:06:33: Started on <hx-noah-31-02>, Execution Home </homes/thiagog'
                  '\n                     enez>, Execution CWD </hps/nobackup/production/ensembl/thi'
                  '\n                     agogenez/pairwises/arabidopsis/run/.>;'
                  '\nThu Mar 18 17:07:47: Resource usage collected.'
                  '\n                     The CPU time used is 53936 seconds.'
                  '\n                     MEM: 344 Mbytes;  SWAP: 1.3 Gbytes;  NTHREAD: 5'
                  '\n                     PGID: 433168;  PIDs: 433168 433177 433179 444026 '
                  '\n'
                  '\n RUNLIMIT                '
                  '\n 10085.0 min'
                  '\n'
                  '\n CORELIMIT MEMLIMIT'
                  '\n      0 M     3.9 G '
                  '\n'
                  '\n MEMORY USAGE:'
                  '\n MAX MEM: 2.5 Gbytes;  AVG MEM: 343 Mbytes'
                  '\n'
                  '\n SCHEDULING PARAMETERS:'
                  '\n           r15s   r1m  r15m   ut      pg    io   ls    it    tmp    swp    mem'
                  '\n loadSched   -     -     -     -    10.0     -    -     -   500M     -   1000M '
                  '\n loadStop    -     -     -     -       -     -    -     -     -      -      -  '
                  '\n'
                  '\n          availcpus '
                  '\n loadSched      1.0 '
                  '\n loadStop        -  '
                  '\n'
                  '\n RESOURCE REQUIREMENT DETAILS:'
                  '\n Combined: select[((mem>4000)) && (type == local)] order[r15s:pg] rusage[mem=40'
                  '\n                     00.00:duration=8h:decay=0,numcpus=1.00:duration=8h:decay=0'
                  '\n                     ] span[hosts=1]'
                  '\n Effective: select[(((mem>4000))) && (type == local)] order[r15s:pg] rusage[mem'
                  '\n                     =4000.00:duration=8h:decay=0,numcpus=1.00:duration=8h:deca'
                  '\n                     y=0] span[hosts=1] '
                  '\n'
                  '\n')
        max_mem, command = parse_mem_and_cmd_from_output(output=output)
        assert len(max_mem.groups()) == 1
        expected_mem = '2.5 Gbytes'
        assert max_mem.group(1) == expected_mem, f'Actual: {max_mem.group(1)}, Expected: "{expected_mem}"'

        assert len(command.groups()) == 1
        expected_command = ('_toil_worker CactusBarRecursion file:/hps/nobackup/production/ensembl/thiagogenez/'
                            'pairwises/arabidopsis/run/jobstore/3 kind-CactusBarRecursion/instance-iu6wo56x '
                            '--context gAShortenedh32xqlE51Yi4=')
        assert command.group(1) == expected_command, f'Actual: {command.group(1)}, Expected: "{expected_command}"'
        print(command)

        max_mem, command = parse_mem_and_cmd_from_output(output='')
        assert max_mem == None
        assert command == None



# Copyright (C) 2015-2021 Regents of the University of California
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import os

from toil.job import Job
from toil.test import ToilTest, get_temp_file
from toil.test.src.jobTest import fn1Test


class JobEncapsulationTest(ToilTest):
    """Tests testing the EncapsulationJob class."""
    def testEncapsulation(self):
        """
        Tests the Job.encapsulation method, which uses the EncapsulationJob
        class.
        """
        # Temporary file
        outFile = get_temp_file(rootDir=self._createTempDir())
        try:
            # Encapsulate a job graph
            a = Job.wrapJobFn(encapsulatedJobFn, "A", outFile, name="a")
            a = a.encapsulate(name="a-encap")
            # Now add children/follow to the encapsulated graph
            d = Job.wrapFn(fn1Test, a.rv(), outFile, name="d")
            e = Job.wrapFn(fn1Test, d.rv(), outFile, name="e")
            a.addChild(d)
            a.addFollowOn(e)
            # Create the runner for the workflow.
            options = Job.Runner.getDefaultOptions(self._getTestJobStorePath())
            options.logLevel = "INFO"
            # Run the workflow, the return value being the number of failed jobs
            Job.Runner.startToil(a, options)
            # Check output
            self.assertEqual(open(outFile).readline(), "ABCDE")
        finally:
            os.remove(outFile)

    def testAddChildEncapsulate(self):
        """
        Make sure that the encapsulate child does not have two parents
        with unique roots.
        """
        # Temporary file
        a = Job.wrapFn(noOp)
        b = Job.wrapFn(noOp)
        a.addChild(b).encapsulate()
        self.assertEqual(len(a.getRootJobs()), 1)


def noOp():
    pass

def encapsulatedJobFn(job, string, outFile):
    a = job.addChildFn(fn1Test, string, outFile, name="inner-a")
    b = a.addFollowOnFn(fn1Test, a.rv(), outFile, name="inner-b")
    return b.rv()



baseVersion = '6.1.0'
cgcloudVersion = '1.6.0a1.dev393'
version = '6.1.0-3f9cba3766e52866ea80d0934498f8c8f3129c3f'
cacheTag = 'cache-local-py3.10'
mainCacheTag = 'cache-master-py3.10'
distVersion = '6.1.0'
exactPython = 'python3.10'
python = 'python3.10'
dockerTag = '6.1.0-3f9cba3766e52866ea80d0934498f8c8f3129c3f-py3.10'
currentCommit = '3f9cba3766e52866ea80d0934498f8c8f3129c3f'
dockerRegistry = 'quay.io/ucsc_cgl'
dockerName = 'toil'
dirty = False
cwltool_version = '3.1.20240112164112'



# Copyright (C) 2015-2021 Regents of the University of California
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from toil.exceptions import FailedJobsException
from toil.job import Job
from toil.jobStores.abstractJobStore import NoSuchFileException
from toil.test import ToilTest, slow


class CheckpointTest(ToilTest):
    def testCheckpointNotRetried(self):
        """A checkpoint job should not be retried if the workflow has a retryCount of 0."""
        options = Job.Runner.getDefaultOptions(self._getTestJobStorePath())
        options.retryCount = 0
        # We set the workflow to *succeed* if the checkpointed job is
        # retried despite the retry count of 0. So, if the workflow
        # fails, this test succeeds.
        # This ends up being a bit confusing, but it's the easiest way
        # to check that the job wasn't retried short of parsing the
        # log.
        with self.assertRaises(FailedJobsException):
            Job.Runner.startToil(CheckRetryCount(numFailuresBeforeSuccess=1), options)

    @slow
    def testCheckpointRetriedOnce(self):
        """A checkpoint job should be retried exactly once if the workflow has a retryCount of 1."""
        options = Job.Runner.getDefaultOptions(self._getTestJobStorePath())
        options.retryCount = 1

        # This should succeed (the checkpoint should be retried once, then succeed)
        try:
            Job.Runner.startToil(CheckRetryCount(numFailuresBeforeSuccess=1), options)
        except FailedJobsException:
            self.fail("The checkpoint job wasn't retried enough times.")

        # This should fail (the checkpoint should be retried once, then fail again)
        with self.assertRaises(FailedJobsException):
            Job.Runner.startToil(CheckRetryCount(numFailuresBeforeSuccess=2), options)

    @slow
    def testCheckpointedRestartSucceeds(self):
        """A checkpointed job should succeed on restart of a failed run if its child job succeeds."""
        options = Job.Runner.getDefaultOptions(self._getTestJobStorePath())
        # The child job should fail the first time
        with self.assertRaises(FailedJobsException):
            Job.Runner.startToil(CheckpointFailsFirstTime(), options)

        # The second time, everything should work
        options.restart = True
        try:
            Job.Runner.startToil(CheckpointFailsFirstTime(), options)
        except FailedJobsException:
            self.fail("Checkpointed workflow restart doesn't clean failures.")

class CheckRetryCount(Job):
    """Fail N times, succeed on the next try."""
    def __init__(self, numFailuresBeforeSuccess):
        super().__init__(checkpoint=True)
        self.numFailuresBeforeSuccess = numFailuresBeforeSuccess

    def getNumRetries(self, fileStore):
        """Mark a retry in the fileStore, and return the number of retries so far."""
        try:
            with fileStore.jobStore.read_shared_file_stream("checkpointRun") as f:
                timesRun = int(f.read().decode('utf-8'))
        except NoSuchFileException:
            timesRun = 0
        with fileStore.jobStore.write_shared_file_stream("checkpointRun") as f:
            f.write(str(timesRun + 1).encode('utf-8'))
        return timesRun

    def run(self, fileStore):
        retryCount = self.getNumRetries(fileStore)
        fileStore.log_to_leader(str(retryCount))
        if retryCount < self.numFailuresBeforeSuccess:
            self.addChild(AlwaysFail())

class AlwaysFail(Job):
    def run(self, fileStore):
        raise RuntimeError(":(")

class CheckpointFailsFirstTime(Job):
    def __init__(self):
        super().__init__(checkpoint=True)

    def run(self, fileStore):
        self.addChild(FailOnce())

class FailOnce(Job):
    """Fail the first time the workflow is run, but succeed thereafter."""
    def run(self, fileStore):
        if fileStore.jobStore.config.workflowAttemptNumber < 1:
            raise RuntimeError("first time around")



# Copyright (C) 2015-2021 Regents of the University of California
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""A demonstration of toil. Sorts the lines of a file into ascending order by doing a parallel merge sort.
"""
import codecs
import os
import random
import shutil

from configargparse import ArgumentParser

from toil.common import Toil
from toil.job import Job
from toil.realtimeLogger import RealtimeLogger

defaultLines = 1000
defaultLineLen = 50
sortMemory = '600M'


def setup(job, inputFile, N, downCheckpoints, options):
    """
    Sets up the sort.
    Returns the FileID of the sorted file
    """
    RealtimeLogger.info("Starting the merge sort")
    return job.addChildJobFn(down,
                             inputFile, N, 'root',
                             downCheckpoints,
                             options = options,
                             preemptible=True,
                             memory=sortMemory).rv()


def down(job, inputFileStoreID, N, path, downCheckpoints, options, memory=sortMemory):
    """
    Input is a file, a subdivision size N, and a path in the hierarchy of jobs.
    If the range is larger than a threshold N the range is divided recursively and
    a follow on job is then created which merges back the results else
    the file is sorted and placed in the output.
    """

    RealtimeLogger.info("Down job starting: %s" % path)

    # Read the file
    inputFile = job.fileStore.readGlobalFile(inputFileStoreID, cache=False)
    length = os.path.getsize(inputFile)
    if length > N:
        # We will subdivide the file
        RealtimeLogger.critical("Splitting file: %s of size: %s"
                % (inputFileStoreID, length))
        # Split the file into two copies
        midPoint = getMidPoint(inputFile, 0, length)
        t1 = job.fileStore.getLocalTempFile()
        with open(t1, 'w') as fH:
            fH.write(copySubRangeOfFile(inputFile, 0, midPoint+1))
        t2 = job.fileStore.getLocalTempFile()
        with open(t2, 'w') as fH:
            fH.write(copySubRangeOfFile(inputFile, midPoint+1, length))
        # Call down recursively. By giving the rv() of the two jobs as inputs to the follow-on job, up,
        # we communicate the dependency without hindering concurrency.
        result = job.addFollowOnJobFn(up,
                                    job.addChildJobFn(down, job.fileStore.writeGlobalFile(t1), N, path + '/0',
                                                      downCheckpoints, checkpoint=downCheckpoints, options=options,
                                                      preemptible=True, memory=options.sortMemory).rv(),
                                    job.addChildJobFn(down, job.fileStore.writeGlobalFile(t2), N, path + '/1',
                                                      downCheckpoints, checkpoint=downCheckpoints, options=options,
                                                      preemptible=True, memory=options.mergeMemory).rv(),
                                    path + '/up', preemptible=True, options=options, memory=options.sortMemory).rv()
    else:
        # We can sort this bit of the file
        RealtimeLogger.critical("Sorting file: %s of size: %s"
                % (inputFileStoreID, length))
        # Sort the copy and write back to the fileStore
        shutil.copyfile(inputFile, inputFile + '.sort')
        sort(inputFile + '.sort')
        result = job.fileStore.writeGlobalFile(inputFile + '.sort')

    RealtimeLogger.info("Down job finished: %s" % path)
    return result


def up(job, inputFileID1, inputFileID2, path, options, memory=sortMemory):
    """
    Merges the two files and places them in the output.
    """

    RealtimeLogger.info("Up job starting: %s" % path)

    with job.fileStore.writeGlobalFileStream() as (fileHandle, outputFileStoreID):
        fileHandle = codecs.getwriter('utf-8')(fileHandle)
        with job.fileStore.readGlobalFileStream(inputFileID1) as inputFileHandle1:
            inputFileHandle1 = codecs.getreader('utf-8')(inputFileHandle1)
            with job.fileStore.readGlobalFileStream(inputFileID2) as inputFileHandle2:
                inputFileHandle2 = codecs.getreader('utf-8')(inputFileHandle2)
                RealtimeLogger.info("Merging %s and %s to %s"
                    % (inputFileID1, inputFileID2, outputFileStoreID))
                merge(inputFileHandle1, inputFileHandle2, fileHandle)
        # Cleanup up the input files - these deletes will occur after the completion is successful.
        job.fileStore.deleteGlobalFile(inputFileID1)
        job.fileStore.deleteGlobalFile(inputFileID2)

        RealtimeLogger.info("Up job finished: %s" % path)

        return outputFileStoreID


def sort(file):
    """Sorts the given file."""
    with open(file) as f:
        lines = f.readlines()

    lines.sort()

    with open(file, 'w') as f:
        for line in lines:
            f.write(line)


def merge(fileHandle1, fileHandle2, outputFileHandle):
    """
    Merges together two files maintaining sorted order.

    All handles must be text-mode streams.
    """
    line2 = fileHandle2.readline()
    for line1 in fileHandle1.readlines():
        while len(line2) != 0 and line2 <= line1:
            outputFileHandle.write(line2)
            line2 = fileHandle2.readline()
        outputFileHandle.write(line1)
    while len(line2) != 0:
        outputFileHandle.write(line2)
        line2 = fileHandle2.readline()


def copySubRangeOfFile(inputFile, fileStart, fileEnd):
    """
    Copies the range (in bytes) between fileStart and fileEnd to the given
    output file handle.
    """
    with open(inputFile) as fileHandle:
        fileHandle.seek(fileStart)
        data = fileHandle.read(fileEnd - fileStart)
        assert len(data) == fileEnd - fileStart
    return data


def getMidPoint(file, fileStart, fileEnd):
    """
    Finds the point in the file to split.
    Returns an int i such that fileStart <= i < fileEnd
    """
    with open(file) as f:
        midPoint = (fileStart + fileEnd) // 2
        assert midPoint >= fileStart
        f.seek(midPoint)
        line = f.readline()
        assert len(line) >= 1
        if len(line) + midPoint < fileEnd:
            return midPoint + len(line) - 1
        f.seek(fileStart)
        line = f.readline()
        assert len(line) >= 1
        assert len(line) + fileStart <= fileEnd
    return len(line) + fileStart - 1


def makeFileToSort(fileName, lines=defaultLines, lineLen=defaultLineLen):
    with open(fileName, 'w') as f:
        for _ in range(lines):
            line = "".join(random.choice('actgACTGNXYZ') for _ in range(lineLen - 1)) + '\n'
            f.write(line)


def main(options=None):
    if not options:
        # deal with command line arguments
        parser = ArgumentParser()
        Job.Runner.addToilOptions(parser)
        parser.add_argument('--numLines', default=defaultLines, help='Number of lines in file to sort.', type=int)
        parser.add_argument('--lineLength', default=defaultLineLen, help='Length of lines in file to sort.', type=int)
        parser.add_argument("--fileToSort", help="The file you wish to sort")
        parser.add_argument("--outputFile", help="Where the sorted output will go")
        parser.add_argument("--overwriteOutput", help="Write over the output file if it already exists.", default=True)
        parser.add_argument("--N", dest="N",
                            help="The threshold below which a serial sort function is used to sort file. "
                                 "All lines must of length less than or equal to N or program will fail",
                            default=10000)
        parser.add_argument('--downCheckpoints', action='store_true',
                            help='If this option is set, the workflow will make checkpoints on its way through'
                                 'the recursive "down" part of the sort')
        parser.add_argument("--sortMemory", dest="sortMemory",
                        help="Memory for jobs that sort chunks of the file.",
                        default=None)

        parser.add_argument("--mergeMemory", dest="mergeMemory",
                        help="Memory for jobs that collate results.",
                        default=None)

        options = parser.parse_args()
    if not hasattr(options, "sortMemory") or not options.sortMemory:
        options.sortMemory = sortMemory
    if not hasattr(options, "mergeMemory") or not options.mergeMemory:
        options.mergeMemory = sortMemory

    # do some input verification
    sortedFileName = options.outputFile or "sortedFile.txt"
    if not options.overwriteOutput and os.path.exists(sortedFileName):
        print(f'Output file {sortedFileName} already exists.  '
              f'Delete it to run the sort example again or use --overwriteOutput=True')
        exit()

    fileName = options.fileToSort
    if options.fileToSort is None:
        # make the file ourselves
        fileName = 'fileToSort.txt'
        if os.path.exists(fileName):
            print(f'Sorting existing file: {fileName}')
        else:
            print(f'No sort file specified. Generating one automatically called: {fileName}.')
            makeFileToSort(fileName=fileName, lines=options.numLines, lineLen=options.lineLength)
    else:
        if not os.path.exists(options.fileToSort):
            raise RuntimeError("File to sort does not exist: %s" % options.fileToSort)

    if int(options.N) <= 0:
        raise RuntimeError("Invalid value of N: %s" % options.N)

    # Now we are ready to run
    with Toil(options) as workflow:
        sortedFileURL = 'file://' + os.path.abspath(sortedFileName)
        if not workflow.options.restart:
            sortFileURL = 'file://' + os.path.abspath(fileName)
            sortFileID = workflow.importFile(sortFileURL)
            sortedFileID = workflow.start(Job.wrapJobFn(setup,
                                                        sortFileID,
                                                        int(options.N),
                                                        options.downCheckpoints,
                                                        options=options,
                                                        memory=sortMemory))
        else:
            sortedFileID = workflow.restart()
        workflow.exportFile(sortedFileID, sortedFileURL)


if __name__ == '__main__':
    main()



# Copyright (C) 2015-2021 Regents of the University of California
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import logging
import mimetypes
import os
import subprocess
import sys

from toil.test import ToilTest, slow
from toil.test.mesos import helloWorld

logging.basicConfig(level=logging.DEBUG)
logger = logging.getLogger(__name__)


class RegularLogTest(ToilTest):

    def setUp(self) -> None:
        super().setUp()
        self.tempDir = self._createTempDir(purpose='tempDir')

    def _getFiles(self, dir):
        return [os.path.join(dir, f) for f in os.listdir(dir) if os.path.isfile(os.path.join(dir, f))]

    def _assertFileTypeExists(self, dir, extension, encoding=None):
        # an encoding of None implies no compression
        logger.info("Checking for %s file in %s", extension, dir)
        onlyFiles = self._getFiles(dir)
        logger.info("Found: %s", str(os.listdir(dir)))
        onlyLogs = [f for f in onlyFiles if f.endswith(extension)]
        logger.info("Found matching: %s", str(onlyLogs))
        assert onlyLogs

        if encoding is not None:
            for log in onlyLogs:
                with open(log, "rb") as f:
                    logger.info("Checking for encoding %s on file %s", str(encoding), log)
                    if encoding == "gzip":
                        # Check for gzip magic header '\x1f\x8b'
                        assert f.read().startswith(b'\x1f\x8b')
                    else:
                        mime = mimetypes.guess_type(log)
                        self.assertEqual(mime[1], encoding)

    @slow
    def testLogToMaster(self):
        toilOutput = subprocess.check_output([sys.executable,
                                              '-m', helloWorld.__name__,
                                              './toilTest',
                                              '--clean=always',
                                              '--logLevel=info'], stderr=subprocess.STDOUT)
        assert helloWorld.childMessage in toilOutput.decode('utf-8')

    def testWriteLogs(self):
        subprocess.check_call([sys.executable,
                               '-m', helloWorld.__name__,
                               './toilTest',
                               '--clean=always',
                               '--logLevel=debug',
                               '--writeLogs=%s' % self.tempDir])
        self._assertFileTypeExists(self.tempDir, '.log')

    @slow
    def testWriteGzipLogs(self):
        subprocess.check_call([sys.executable,
                               '-m', helloWorld.__name__,
                               './toilTest',
                               '--clean=always',
                               '--logLevel=debug',
                               '--writeLogsGzip=%s' % self.tempDir])
        self._assertFileTypeExists(self.tempDir, '.log.gz', 'gzip')

    @slow
    def testMultipleLogToMaster(self):
        toilOutput = subprocess.check_output([sys.executable,
                                              '-m', helloWorld.__name__,
                                              './toilTest',
                                              '--clean=always',
                                              '--logLevel=info'], stderr=subprocess.STDOUT)
        assert helloWorld.parentMessage in toilOutput.decode('utf-8')

    def testRegularLog(self):
        toilOutput = subprocess.check_output([sys.executable,
                                              '-m', helloWorld.__name__,
                                              './toilTest',
                                              '--clean=always',
                                              '--batchSystem=single_machine',
                                              '--logLevel=debug'], stderr=subprocess.STDOUT)
        assert "single machine batch system" in toilOutput.decode('utf-8')



# Copyright (C) 2015-2021 Regents of the University of California
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import errno
import logging
import os
import random

from toil.common import Toil
from toil.fileStores import FileID
from toil.job import Job
from toil.test import ToilTest, slow

logger = logging.getLogger(__name__)

PREFIX_LENGTH=200


# TODO: This test is ancient and while similar tests exist in `fileStoreTest.py`, none of them look
# at the contents of read files and thus we will let this test remain as-is.
class JobFileStoreTest(ToilTest):
    """
    Tests testing the methods defined in :class:toil.fileStores.abstractFileStore.AbstractFileStore.
    """
    def testCachingFileStore(self):
        options = Job.Runner.getDefaultOptions(self._getTestJobStorePath())
        with Toil(options) as workflow:
            workflow.start(Job.wrapJobFn(simpleFileStoreJob))

    def testNonCachingFileStore(self):
        options = Job.Runner.getDefaultOptions(self._getTestJobStorePath())
        options.caching = False
        with Toil(options) as workflow:
            workflow.start(Job.wrapJobFn(simpleFileStoreJob))

    def _testJobFileStore(self, retryCount=0, badWorker=0.0, stringNo=1, stringLength=1000000,
                          testNo=2):
        """
        Creates a chain of jobs, each reading and writing files using the
        toil.fileStores.abstractFileStore.AbstractFileStore interface. Verifies the files written are always what we
        expect.
        """
        for test in range(testNo):
            #Make a list of random strings, each of 100k chars and hash the first 200
            #base prefix to the string
            def randomString():
                chars = "ABCDEFGHIJKLMNOPQRSTUVWXYZ"
                s = "".join([random.choice(chars) for i in range(stringLength)])
                return s[:PREFIX_LENGTH], s
            #Total length is 2 million characters (20 strings of length 100K each)
            testStrings = dict([randomString() for i in range(stringNo)])
            options = Job.Runner.getDefaultOptions(self._getTestJobStorePath())
            options.logLevel = "DEBUG"
            options.retryCount=retryCount
            options.badWorker=badWorker
            options.badWorkerFailInterval = 1.0
            chainLength = 10
            # Run the workflow, the return value being the number of failed jobs
            Job.Runner.startToil(Job.wrapJobFn(fileTestJob, [],
                                               testStrings, chainLength),
                                 options)

    def testJobFileStore(self):
        """
        Tests case that about half the files are cached
        """
        self._testJobFileStore(retryCount=0, badWorker=0.0,  stringNo=5, stringLength=1000000)

    @slow
    def testJobFileStoreWithBadWorker(self):
        """
        Tests case that about half the files are cached and the worker is randomly
        failing.
        """
        self._testJobFileStore(retryCount=100, badWorker=0.5,  stringNo=5, stringLength=1000000)


def fileTestJob(job, inputFileStoreIDs, testStrings, chainLength):
    """
    Test job exercises toil.fileStores.abstractFileStore.AbstractFileStore functions
    """
    outputFileStoreIds = [] #Strings passed to the next job in the chain

    #Load the input jobStoreFileIDs and check that they map to the
    #same set of random input strings, exercising the different functions in the fileStore interface
    for fileStoreID in inputFileStoreIDs:
        if random.random() > 0.5:
            #Read the file for the fileStoreID, randomly picking a way to invoke readGlobalFile
            if random.random() > 0.5:
                local_path = job.fileStore.getLocalTempFileName() if random.random() > 0.5 else None
                cache = random.random() > 0.5

                tempFile = job.fileStore.readGlobalFile(fileStoreID,
                                                        local_path,
                                                        cache=cache)
                with open(tempFile) as fH:
                    string = fH.readline()
                logging.info("Downloaded %s to local path %s with cache %s and got %s with %d letters",
                              fileStoreID, local_path, cache, tempFile, len(string))
            else:
                #Check the local file is as we expect
                with job.fileStore.readGlobalFileStream(fileStoreID, 'utf-8') as fH:
                    string = fH.readline()
                logging.info("Streamed %s and got %d letters", fileStoreID, len(string))
            #Check the string we get back is what we expect
            assert string[:PREFIX_LENGTH] in testStrings, f"Could not find string: {string[:PREFIX_LENGTH]}"
            assert testStrings[string[:PREFIX_LENGTH]] == string, f"Mismatch in string: {string[:PREFIX_LENGTH]}"

            #This allows the file to be passed to the next job
            outputFileStoreIds.append(fileStoreID)
        else:
            #This tests deletion
            logging.info("Deleted %s", fileStoreID)
            job.fileStore.deleteGlobalFile(fileStoreID)

    #Fill out the output strings until we have the same number as the input strings
    #exercising different ways of writing files to the file store
    while len(outputFileStoreIds) < len(testStrings):
        #Pick a string and write it into a file
        testString = random.choice(list(testStrings.values()))
        if random.random() > 0.5:
            #Make a local copy of the file
            tempFile = job.fileStore.getLocalTempFile() if random.random() > 0.5 \
            else os.path.join(job.fileStore.getLocalTempDir(), "temp.txt")
            with open(tempFile, 'w') as fH:
                fH.write(testString)
            #Write a local copy of the file using the local file
            fileStoreID = job.fileStore.writeGlobalFile(tempFile)

            # Make sure it returned a valid and correct FileID with the right size
            assert isinstance(fileStoreID, FileID)
            assert fileStoreID.size == len(testString.encode('utf-8'))

            outputFileStoreIds.append(fileStoreID)
        else:
            #Use the writeGlobalFileStream method to write the file
            with job.fileStore.writeGlobalFileStream() as (fH, fileStoreID):
                fH.write(testString.encode('utf-8'))
                outputFileStoreIds.append(fileStoreID)

            #Make sure it returned a valid and correct FileID with the right size
            assert isinstance(fileStoreID, FileID)
            assert fileStoreID.size == len(testString.encode('utf-8'))

    if chainLength > 0:
        #Make a child that will read these files and check it gets the same results
        job.addChildJobFn(fileTestJob, outputFileStoreIds, testStrings, chainLength-1)


fileStoreString = "Testing writeGlobalFile"
streamingFileStoreString = "Testing writeGlobalFileStream"


def simpleFileStoreJob(job):
    localFilePath = os.path.join(job.fileStore.getLocalTempDir(), "parentTemp.txt")
    with open(localFilePath, 'w') as f:
        f.write(fileStoreString)
    testID1 = job.fileStore.writeGlobalFile(localFilePath)

    testID2 = None
    with job.fileStore.writeGlobalFileStream() as (f, fileID):
        f.write(streamingFileStoreString.encode('utf-8'))
        testID2 = fileID

    job.addChildJobFn(fileStoreChild, testID1, testID2)


def fileStoreChild(job, testID1, testID2):
    with job.fileStore.readGlobalFileStream(testID1) as f:
        assert(f.read().decode('utf-8') == fileStoreString)

    localFilePath = os.path.join(job.fileStore.getLocalTempDir(), "childTemp.txt")
    job.fileStore.readGlobalFile(testID2, localFilePath)
    with open(localFilePath) as f:
        assert(f.read() == streamingFileStoreString)

    job.fileStore.deleteLocalFile(testID2)
    try:
        job.fileStore.deleteLocalFile(testID1)
    except OSError as e:
        if e.errno == errno.ENOENT:  # indicates that the file was not found
            pass
        else:
            raise
    else:
        raise RuntimeError("Deleting a non-existant file did not throw an exception")

    for fileID in (testID1, testID2):
        job.fileStore.deleteGlobalFile(fileID)



# Copyright (C) 2015-2021 Regents of the University of California
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import fcntl
import logging
import os
import tempfile
from collections import namedtuple
from contextlib import contextmanager

import dill

from toil.lib.io import robust_rmtree
from toil.realtimeLogger import RealtimeLogger
from toil.resource import ModuleDescriptor

logger = logging.getLogger(__name__)


class DeferredFunction(namedtuple('DeferredFunction', 'function args kwargs name module')):
    """
    >>> from collections import defaultdict
    >>> df = DeferredFunction.create(defaultdict, None, {'x':1}, y=2)
    >>> df
    DeferredFunction(defaultdict, ...)
    >>> df.invoke() == defaultdict(None, x=1, y=2)
    True
    """
    @classmethod
    def create(cls, function, *args, **kwargs):
        """
        Capture the given callable and arguments as an instance of this class.

        :param callable function: The deferred action to take in the form of a function
        :param tuple args: Non-keyword arguments to the function
        :param dict kwargs: Keyword arguments to the function
        """
        # The general principle is to deserialize as late as possible, i.e. when the function is
        # to be invoked, as that will avoid redundantly deserializing deferred functions for
        # concurrently running jobs when the cache state is loaded from disk. By implication we
        # should serialize as early as possible. We need to serialize the function as well as its
        # arguments.
        return cls(*list(map(dill.dumps, (function, args, kwargs))),
                   name=function.__name__,
                   module=ModuleDescriptor.forModule(function.__module__).globalize())

    def invoke(self):
        """
        Invoke the captured function with the captured arguments.
        """
        logger.debug('Running deferred function %s.', self)
        self.module.makeLoadable()
        function, args, kwargs = list(map(dill.loads, (self.function, self.args, self.kwargs)))
        return function(*args, **kwargs)

    def __str__(self):
        return f'{self.__class__.__name__}({self.name}, ...)'

    __repr__ = __str__


class DeferredFunctionManager:
    """
    Implements a deferred function system. Each Toil worker will have an
    instance of this class. When a job is executed, it will happen inside a
    context manager from this class. If the job registers any "deferred"
    functions, they will be executed when the context manager is exited.

    If the Python process terminates before properly exiting the context
    manager and running the deferred functions, and some other worker process
    enters or exits the per-job context manager of this class at a later time,
    or when the DeferredFunctionManager is shut down on the worker, the earlier
    job's deferred functions will be picked up and run.

    Note that deferred function cleanup is on a best-effort basis, and deferred
    functions may end up getting executed multiple times.

    Internally, deferred functions are serialized into files in the given
    directory, which are locked by the owning process.

    If that process dies, other processes can detect that the files are able to
    be locked, and will take them over.
    """

    # Define what directory the state directory should actaully be, under the base
    STATE_DIR_STEM = 'deferred'
    # Have a prefix to distinguish our deferred functions from e.g. NFS
    # "silly rename" files, or other garbage that people put in our
    # directory
    PREFIX = 'func'
    # And a suffix to distinguish in-progress from completed files
    WIP_SUFFIX = '.tmp'

    def __init__(self, stateDirBase: str) -> None:
        """
        Create a new DeferredFunctionManager, sharing state with other
        instances in other processes using the given shared state directory.

        Uses a fixed path under that directory for state files. Creates it if
        not present.

        Note that if the current umask lets other people create files in that
        new directory, we are going to execute their code!

        The created directory will be left behind, because we never know if
        another worker will come along later on this node.
        """

        # Work out where state files live
        self.stateDir = os.path.join(stateDirBase, self.STATE_DIR_STEM)
        os.makedirs(self.stateDir, exist_ok=True)

        # We need to get a state file, locked by us and not somebody scanning for abandoned state files.
        # So we suffix not-yet-ready ones with our suffix
        self.stateFD, self.stateFileName = tempfile.mkstemp(dir=self.stateDir,
                                                            prefix=self.PREFIX,
                                                            suffix=self.WIP_SUFFIX)

        # Lock the state file. The lock will automatically go away if our process does.
        try:
            fcntl.lockf(self.stateFD, fcntl.LOCK_EX | fcntl.LOCK_NB)
        except OSError as e:
            # Someone else might have locked it even though they should not have.
            raise RuntimeError(f"Could not lock deferred function state file {self.stateFileName}: {str(e)}")

        # Rename it to remove the suffix
        os.rename(self.stateFileName, self.stateFileName[:-len(self.WIP_SUFFIX)])
        self.stateFileName = self.stateFileName[:-len(self.WIP_SUFFIX)]

        # Wrap the FD in a Python file object, which we will use to actually use it.
        # Problem: we can't be readable and writable at the same time. So we need two file objects.
        self.stateFileOut = os.fdopen(self.stateFD, 'wb')
        self.stateFileIn = open(self.stateFileName, 'rb')

        logger.debug("Opened with own state file %s" % self.stateFileName)

    def __del__(self):
        """
        Clean up our state on disk. We assume that the deferred functions we
        manage have all been executed, and none are currently recorded.
        """

        logger.debug("Removing own state file %s" % self.stateFileName)

        # Hide the state from other processes
        if os.path.exists(self.stateFileName):
            os.unlink(self.stateFileName)

        # Unlock it
        fcntl.lockf(self.stateFD, fcntl.LOCK_UN)

        # Don't bother with close, destroying will close and it seems to maybe
        # have been GC'd already anyway.

    @contextmanager
    def open(self):
        """
        Yields a single-argument function that allows for deferred functions of
        type :class:`toil.DeferredFunction` to be registered.  We use this
        design so deferred functions can be registered only inside this context
        manager.

        Not thread safe.
        """

        # Clean up other jobs before we run, so our job has a nice clean node
        self._runOrphanedDeferredFunctions()

        try:
            def defer(deferredFunction):
                # Just serialize defered functions one after the other.
                # If serializing later ones fails, eariler ones will still be intact.
                # We trust dill to protect sufficiently against partial reads later.
                logger.debug("Deferring function %s" % repr(deferredFunction))
                dill.dump(deferredFunction, self.stateFileOut)
                # Flush before returning so we can guarantee the write is on disk if we die.
                self.stateFileOut.flush()

            logger.debug("Running job")
            yield defer
        finally:
            self._runOwnDeferredFunctions()
            self._runOrphanedDeferredFunctions()

    @classmethod
    def cleanupWorker(cls, stateDirBase: str) -> None:
        """
        Called by the batch system when it shuts down the node, after all
        workers are done, if the batch system supports worker cleanup. Checks
        once more for orphaned deferred functions and runs them.
        """

        logger.debug("Cleaning up deferred functions system")

        # Open up
        cleaner = cls(stateDirBase)
        # Do the final round of cleanup
        cleaner._runOrphanedDeferredFunctions()

        # Close all the files in there.
        del cleaner

        try:
            robust_rmtree(os.path.join(stateDirBase, cls.STATE_DIR_STEM))
        except OSError as err:
            logger.exception(err)
            # we tried, lets move on


    def _runDeferredFunction(self, deferredFunction):
        """
        Run a deferred function (either our own or someone else's).

        Reports an error if it fails.
        """

        try:
            deferredFunction.invoke()
        except Exception as err:
            # Report this in real time, if enabled. Otherwise the only place it ends up is the worker log.
            RealtimeLogger.error("Failed to run deferred function %s: %s", repr(deferredFunction), str(err))
        except:
            RealtimeLogger.error("Failed to run deferred function %s", repr(deferredFunction))

    def _runAllDeferredFunctions(self, fileObj):
        """
        Read and run deferred functions until EOF from the given open file.
        """

        try:
            while True:
                # Load each function
                deferredFunction = dill.load(fileObj)
                logger.debug("Loaded deferred function %s" % repr(deferredFunction))
                # Run it
                self._runDeferredFunction(deferredFunction)
        except EOFError as e:
            # This is expected and means we read all the complete entries.
            logger.debug("Out of deferred functions!")

    def _runOwnDeferredFunctions(self):
        """
        Run all of the deferred functions that were registered.
        """

        logger.debug("Running own deferred functions")

        # Seek back to the start of our file
        self.stateFileIn.seek(0)

        # Read and run each function in turn
        self._runAllDeferredFunctions(self.stateFileIn)

        # Go back to the beginning and truncate, to prepare for a new set of deferred functions.
        self.stateFileIn.seek(0)
        self.stateFileOut.seek(0)
        self.stateFileOut.truncate()

    def _runOrphanedDeferredFunctions(self):
        """
        Scan for files that aren't locked by anybody and run all their deferred functions, then clean them up.
        """

        logger.debug("Running orphaned deferred functions")
        states_handled = 0

        # Track whether we found any work to do.
        # We will keep looping as long as there is work to do.
        foundFiles = True

        while foundFiles:
            # Clear this out unless we find some work we can get ahold of.
            foundFiles = False

            for filename in os.listdir(self.stateDir):
                # Scan the whole directory for work nobody else owns.

                if filename.endswith(self.WIP_SUFFIX):
                    # Skip files from instances that are still being set up
                    continue

                if not filename.startswith(self.PREFIX):
                    # Skip NFS deleted files and any other contaminants
                    continue

                fullFilename = os.path.join(self.stateDir, filename)

                if fullFilename == self.stateFileName:
                    # We would be able to lock our own file, and it would appear unowned.
                    # So skip it.
                    continue

                # We need to make sure that we don't hold two
                # DeferredFunctionManagers at once! So make sure to del yours
                # when you are done with it. TODO: Make it a singleton!

                fd = None

                try:
                    # Try locking each file.
                    # The file may have vanished since we saw it, so we have to ignore failures.
                    # We open in read write mode because the fcntl docs say you
                    # might only be able to exclusively lock files opened for
                    # writing.
                    fd = os.open(fullFilename, os.O_RDWR)
                except OSError:
                    # Maybe the file vanished. Try the next one
                    continue

                try:
                    fcntl.lockf(fd, fcntl.LOCK_EX | fcntl.LOCK_NB)
                except OSError:
                    # File is still locked by someone else.
                    # Look at the next file instead
                    continue

                logger.debug("Locked file %s" % fullFilename)

                # File is locked successfully. Our problem now.
                foundFiles = True

                # Actually run all the stored deferred functions
                fileObj = os.fdopen(fd, 'rb')
                self._runAllDeferredFunctions(fileObj)
                states_handled += 1

                try:
                    # Ok we are done with this file. Get rid of it so nobody else does it.
                    os.unlink(fullFilename)
                except OSError:
                    # Maybe the file vanished.
                    pass

                # Unlock it
                fcntl.lockf(fd, fcntl.LOCK_UN)

                # Now close it. This closes the backing file descriptor. See
                # <https://stackoverflow.com/a/24984929>
                fileObj.close()

        logger.debug("Ran orphaned deferred functions from %d abandoned state files", states_handled)



# Copyright (C) 2015-2021 Regents of the University of California
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import unittest

from docker.errors import ImageNotFound
from toil import checkDockerImageExists, parseDockerAppliance
from toil.test import ToilTest, needs_docker


@needs_docker
class DockerCheckTest(ToilTest):
    """Tests checking whether a docker image exists or not."""

    @unittest.skip('Consumes unauthenticated Docker Hub pulls if run')
    def testOfficialUbuntuRepo(self):
        """Image exists.  This should pass."""
        ubuntu_repo = 'ubuntu:latest'
        assert checkDockerImageExists(ubuntu_repo)

    @unittest.skip('Consumes unauthenticated Docker Hub pulls if run')
    def testBroadDockerRepo(self):
        """Image exists.  This should pass."""
        broad_repo = 'broadinstitute/genomes-in-the-cloud:2.0.0'
        assert checkDockerImageExists(broad_repo)

    @unittest.skip('Consumes unauthenticated Docker Hub pulls if run')
    def testBroadDockerRepoBadTag(self):
        """Bad tag.  This should raise."""
        broad_repo = 'broadinstitute/genomes-in-the-cloud:-----'
        with self.assertRaises(ImageNotFound):
            checkDockerImageExists(broad_repo)

    @unittest.skip('Consumes unauthenticated Docker Hub pulls if run')
    def testNonexistentRepo(self):
        """Bad image.  This should raise."""
        nonexistent_repo = '------:-----'
        with self.assertRaises(ImageNotFound):
            checkDockerImageExists(nonexistent_repo)

    def testToilQuayRepo(self):
        """Image exists.  Should pass."""
        toil_repo = 'quay.io/ucsc_cgl/toil:latest'
        assert checkDockerImageExists(toil_repo)

    def testBadQuayRepoNTag(self):
        """Bad repo and tag.  This should raise."""
        nonexistent_quay_repo = 'quay.io/--------:---'
        with self.assertRaises(ImageNotFound):
            checkDockerImageExists(nonexistent_quay_repo)

    def testBadQuayRepo(self):
        """Bad repo.  This should raise."""
        nonexistent_quay_repo = 'quay.io/--------:latest'
        with self.assertRaises(ImageNotFound):
            checkDockerImageExists(nonexistent_quay_repo)

    def testBadQuayTag(self):
        """Bad tag.  This should raise."""
        nonexistent_quay_repo = 'quay.io/ucsc_cgl/toil:---'
        with self.assertRaises(ImageNotFound):
            checkDockerImageExists(nonexistent_quay_repo)

    def testGoogleRepo(self):
        """Image exists.  Should pass."""
        google_repo = 'gcr.io/google-containers/busybox:latest'
        assert checkDockerImageExists(google_repo)

    def testBadGoogleRepo(self):
        """Bad repo and tag.  This should raise."""
        nonexistent_google_repo = 'gcr.io/google-containers/--------:---'
        with self.assertRaises(ImageNotFound):
            checkDockerImageExists(nonexistent_google_repo)

    def testApplianceParser(self):
        """Test that a specified appliance is parsed correctly."""
        docker_list = ['ubuntu:latest',
                       'ubuntu',
                       'broadinstitute/genomes-in-the-cloud:2.0.0',
                       'quay.io/ucsc_cgl/toil:latest',
                       'gcr.io/google-containers/busybox:latest']
        parsings = []
        for image in docker_list:
            registryName, imageName, tag = parseDockerAppliance(image)
            parsings.append([registryName, imageName, tag])
        expected_parsings = [['docker.io', 'ubuntu', 'latest'],
                             ['docker.io', 'ubuntu', 'latest'],
                             ['docker.io', 'broadinstitute/genomes-in-the-cloud', '2.0.0'],
                             ['quay.io', 'ucsc_cgl/toil', 'latest'],
                             ['gcr.io', 'google-containers/busybox', 'latest']]
        assert parsings == expected_parsings



# Copyright (C) 2015-2021 Regents of the University of California
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Reports statistical data about a given Toil workflow."""
import json
import logging
import math
import sys
from argparse import ArgumentParser, Namespace
from functools import partial
from typing import Any, Callable, Dict, List, Optional, TextIO, Union

from toil.common import Config, Toil, parser_with_common_options
from toil.job import Job
from toil.jobStores.abstractJobStore import AbstractJobStore, NoSuchJobStoreException
from toil.lib.expando import Expando
from toil.options.common import SYS_MAX_SIZE
from toil.statsAndLogging import set_logging_from_options

logger = logging.getLogger(__name__)

# These categories of stat will be reported
CATEGORIES = ["time", "clock", "wait", "memory", "disk"]
# These are the units they are stored in
CATEGORY_UNITS = {
    "time": "s",
    "clock": "core-s",
    "wait": "core-s",
    "memory": "KiB",
    "disk": "B"
}
# These are what we call them to the user
TITLES = {
    "time": "Real Time",
    "clock": "CPU Time",
    "wait": "CPU Wait",
    "memory": "Memory",
    "disk": "Disk"
}

# Of those, these are in time
TIME_CATEGORIES = {"time", "clock", "wait"}
# And these are in space
SPACE_CATEGORIES = {"memory", "disk"}
# These categories aren't stored and need to be computed
COMPUTED_CATEGORIES = {"wait"}

# The different kinds of summaries have both short and long names, and we need
# to convert between them.
LONG_FORMS = {
    "med": "median",
    "ave": "average",
    "min": "min",
    "total": "total",
    "max": "max",
}

class ColumnWidths:
    """
    Convenience object that stores the width of columns for printing. Helps make things pretty.
    """

    def __init__(self) -> None:
        self.categories = CATEGORIES
        self.fields_count = ["count", "min", "med", "ave", "max", "total"]
        self.fields = ["min", "med", "ave", "max", "total"]
        self.data: Dict[str, int] = {}
        for category in self.categories:
            for field in self.fields_count:
                self.set_width(category, field, 8)

    def title(self, category: str) -> int:
        """Return the total printed length of this category item."""
        return sum(self.get_width(category, x) for x in self.fields)

    def get_width(self, category: str, field: str) -> int:
        category = category.lower()
        return self.data[f"{category}_{field}"]

    def set_width(self, category: str, field: str, width: int) -> None:
        category = category.lower()
        self.data[f"{category}_{field}"] = width

    def report(self) -> None:
        for c in self.categories:
            for f in self.fields:
                print("%s %s %d" % (c, f, self.get_width(c, f)))


def pad_str(s: str, field: Optional[int] = None) -> str:
    """Pad the beginning of a string with spaces, if necessary."""
    if field is None or len(s) >= field:
        return s
    else:
        return " " * (field - len(s)) + s


def pretty_space(k: float, field: Optional[int] = None, alone: bool = False) -> str:
    """Given input k as kibibytes, return a nicely formatted string."""
    # If we don't have a header to say bytes, include the B.
    trailer = "B" if alone else ""
    if k < 1024:
        return pad_str("%gKi%s" % (k, trailer), field)
    if k < (1024 * 1024):
        return pad_str("%.1fMi%s" % (k / 1024.0, trailer), field)
    if k < (1024 * 1024 * 1024):
        return pad_str("%.1fGi%s" % (k / 1024.0 / 1024.0, trailer), field)
    if k < (1024 * 1024 * 1024 * 1024):
        return pad_str("%.1fTi%s" % (k / 1024.0 / 1024.0 / 1024.0, trailer), field)
    if k < (1024 * 1024 * 1024 * 1024 * 1024):
        return pad_str("%.1fPi%s" % (k / 1024.0 / 1024.0 / 1024.0 / 1024.0, trailer), field)

    # due to https://stackoverflow.com/questions/47149154
    assert False


def pretty_time(t: float, field: Optional[int] = None, unit: str = "s", alone: bool = False) -> str:
    """
    Given input t as seconds, return a nicely formatted string.
    """
    assert unit in ("s", "core-s")
    # Qualify our CPU times as CPU time if we aren't in a table that does that
    unit_str = report_unit(unit) if alone else "s"

    from math import floor

    pluralDict = {True: "s", False: ""}
    if t < 120:
        return pad_str("%d%s" % (t, unit_str), field)
    if t < 120 * 60:
        m = floor(t / 60.0)
        s = t % 60
        return pad_str("%dm%d%s" % (m, s, unit_str), field)
    if t < 25 * 60 * 60:
        h = floor(t / 60.0 / 60.0)
        m = floor((t - (h * 60.0 * 60.0)) / 60.0)
        s = t % 60
        return pad_str("%dh%gm%d%s" % (h, m, s, unit_str), field)
    if t < 7 * 24 * 60 * 60:
        d = floor(t / 24.0 / 60.0 / 60.0)
        h = floor((t - (d * 24.0 * 60.0 * 60.0)) / 60.0 / 60.0)
        m = floor((t - (d * 24.0 * 60.0 * 60.0) - (h * 60.0 * 60.0)) / 60.0)
        s = t % 60
        dPlural = pluralDict[d > 1]
        return pad_str("%dday%s%dh%dm%d%s" % (d, dPlural, h, m, s, unit_str), field)
    w = floor(t / 7.0 / 24.0 / 60.0 / 60.0)
    d = floor((t - (w * 7 * 24 * 60 * 60)) / 24.0 / 60.0 / 60.0)
    h = floor(
        (t - (w * 7.0 * 24.0 * 60.0 * 60.0) - (d * 24.0 * 60.0 * 60.0)) / 60.0 / 60.0
    )
    m = floor(
        (
            t
            - (w * 7.0 * 24.0 * 60.0 * 60.0)
            - (d * 24.0 * 60.0 * 60.0)
            - (h * 60.0 * 60.0)
        )
        / 60.0
    )
    s = t % 60
    wPlural = pluralDict[w > 1]
    dPlural = pluralDict[d > 1]
    return pad_str("%dweek%s%dday%s%dh%dm%d%s" % (w, wPlural, d, dPlural, h, m, s, unit_str), field)

def report_unit(unit: str) -> str:
    """
    Format a unit name for display.
    """
    if unit == "core-s":
        return "core·s"
    return unit

def report_time(t: float, options: Namespace, field: Optional[int] = None, unit: str = "s", alone: bool = False) -> str:
    """Given t seconds, report back the correct format as string."""
    assert unit in ("s", "core-s")
    if options.pretty:
        return pretty_time(t, field=field, unit=unit, alone=alone)
    unit_text = f" {report_unit(unit)}" if alone else ""
    if field is not None:
        assert field >= len(unit_text)
        return "%*.2f%s" % (field - len(unit_text), t, unit_text)
    return "%.2f%s" % (t, unit_text)


def report_space(
        k: float, options: Namespace, field: Optional[int] = None, unit: str = "KiB", alone: bool = False
) -> str:
    """
    Given k kibibytes, report back the correct format as string.

    If unit is set to B, convert to KiB first.
    """
    if unit == "B":
        k /= 1024.0
        unit = "KiB"
    assert unit == "KiB"
    if options.pretty:
        return pretty_space(int(k), field=field, alone=alone)
    else:
        # If we don't have a heading to say bytes, include the B
        trailer = "KiB" if alone else "Ki"
        if field is not None:
            assert field >= len(trailer)
            return "%*d%s" % (field - len(trailer), k, trailer)
        else:
            return "%d%s" % (int(k), trailer)


def report_number(n: Union[int, float, None], field: Optional[int] = None, nan_value: str = "NaN") -> str:
    """
    Given a number, report back the correct format as string.
    
    If it is a NaN or None, use nan_value to represent it instead.
    """
    if n is None or math.isnan(n):
        return pad_str(nan_value, field=field)
    else:
        # Make sure not to format with too much precision for the field size;
        # leave room for . and the spacing to the previous field.
        return "%*.*g" % (field, field - 2, n) if field else "%g" % n

def report(v: float, category: str, options: Namespace, field: Optional[int] = None, alone=False) -> str:
    """
    Report a value of the given category formatted as a string.

    Uses the given field width if set.

    If alone is set, the field is being formatted outside a table and might need a unit.
    """

    unit = CATEGORY_UNITS.get(category)
    if unit in ("s", "core-s"):
        # This is time.
        return report_time(v, options, field=field, unit=unit, alone=alone)
    elif unit in ("B", "KiB"):
        # This is space.
        return report_space(v, options, field=field, unit=unit, alone=alone)
    else:
        raise ValueError(f"Unimplemented unit {unit} for category {category}")

def sprint_tag(
    key: str,
    tag: Expando,
    options: Namespace,
    columnWidths: Optional[ColumnWidths] = None,
) -> str:
    """Generate a pretty-print ready string from a JTTag()."""
    if columnWidths is None:
        columnWidths = ColumnWidths()
    header = "  %7s " % decorate_title("count", "Count", options)
    sub_header = "  %7s " % "n"
    tag_str = f"  {report_number(n=tag.total_number, field=7)}"
    out_str = ""
    if key == "job":
        out_str += " {:<12} | {:>7}{:>7}{:>7}{:>7}\n".format(
            "Worker Jobs", "min", "med", "ave", "max"
        )
        worker_str = "%s| " % (" " * 14)
        for t in [
            tag.min_number_per_worker,
            tag.median_number_per_worker,
            tag.average_number_per_worker,
            tag.max_number_per_worker,
        ]:
            worker_str += report_number(n=t, field=7)
        out_str += worker_str + "\n"

    for category in CATEGORIES:
        if category not in options.categories:
            continue

        header += "| %*s " % (
            columnWidths.title(category),
            decorate_title(category, TITLES[category], options),
        )
        sub_header += decorate_subheader(category, columnWidths, options)
        tag_str += " | "

        for field in ["min", "med", "ave", "max", "total"]:
            t = getattr(tag, f"{LONG_FORMS[field]}_{category}")
            width = columnWidths.get_width(category, field)
            s = report(t, category, options, field=width)
            tag_str += s

    out_str += header + "\n"
    out_str += sub_header + "\n"
    out_str += tag_str + "\n"
    return out_str


def decorate_title(category: str, title: str, options: Namespace) -> str:
    """
    Add extra parts to the category titles.

    Add units to title if they won't appear in the formatted values.
    Add a marker to TITLE if the TITLE is sorted on.
    """
    unit = CATEGORY_UNITS.get(category)
    if unit in ("s", "core-s") and not options.pretty:
        # This is a time and we won't write it out as text, so add a unit.
        title = f"{title} ({report_unit(unit)})"
    elif unit == "core-s" and options.pretty:
        # This is a core-second category and we won't be putting the core unit
        # in the value, so note that here.
        title = f"{title} (core)"
    elif unit in ("B", "KiB"):
        # The Ki part will appear in the cell so we need a B
        title = f"{title} (B)"
    if category.lower() == options.sortCategory:
        return "%s*" % title
    else:
        return title


def decorate_subheader(
    category: str, columnWidths: ColumnWidths, options: Namespace
) -> str:
    """Add a marker to the correct field if the TITLE is sorted on."""
    if category != options.sortCategory:
        s = "| %*s%*s%*s%*s%*s " % (
            columnWidths.get_width(category, "min"),
            "min",
            columnWidths.get_width(category, "med"),
            "med",
            columnWidths.get_width(category, "ave"),
            "ave",
            columnWidths.get_width(category, "max"),
            "max",
            columnWidths.get_width(category, "total"),
            "total",
        )
        return s
    else:
        s = "| "
        for field, width in [
            ("min", columnWidths.get_width(category, "min")),
            ("med", columnWidths.get_width(category, "med")),
            ("ave", columnWidths.get_width(category, "ave")),
            ("max", columnWidths.get_width(category, "max")),
            ("total", columnWidths.get_width(category, "total")),
        ]:
            if options.sortField == field:
                s += "%*s*" % (width - 1, field)
            else:
                s += "%*s" % (width, field)
        s += " "
        return s


def get(tree: Expando, name: str) -> float:
    """Return a float value attribute NAME from TREE."""
    try:
        return float(tree.get(name, "nan"))
    except ValueError:
        return float("nan")


def sort_jobs(jobTypes: List[Any], options: Namespace) -> List[Any]:
    """Return a jobTypes all sorted."""
    sortField = LONG_FORMS[options.sortField]
    if (
        options.sortCategory in CATEGORIES
    ):
        return sorted(
            jobTypes,
            key=lambda tag: getattr(tag, "%s_%s" % (sortField, options.sortCategory)),
            reverse=options.sort == "decending",
        )
    elif options.sortCategory == "alpha":
        return sorted(
            jobTypes,
            key=lambda tag: tag.name,  # type: ignore
            reverse=options.sort == "decending",
        )
    elif options.sortCategory == "count":
        return sorted(
            jobTypes,
            key=lambda tag: tag.total_number,  # type: ignore
            reverse=options.sort == "decending",
        )

    # due to https://stackoverflow.com/questions/47149154
    assert False


def report_pretty_data(
    root: Expando,
    worker: Expando,
    job: Expando,
    job_types: List[Any],
    options: Namespace,
) -> str:
    """Print the important bits out."""
    out_str = "Batch System: %s\n" % root.batch_system
    out_str += "Default Cores: %s  Default Memory: %s\n" "Max Cores: %s\n" % (
        report_number(n=get(root, "default_cores")),
        # Although per-job memory usage is in KiB, our default is stored in bytes.
        report_space(get(root, "default_memory"), options, unit="B", alone=True),
        report_number(n=get(root, "max_cores"), nan_value="unlimited"),
    )
    out_str += "Local CPU Time: {}  Overall Runtime: {}\n".format(
        report(get(root, "total_clock"), "clock", options, alone=True),
        report(get(root, "total_run_time"), "time", options, alone=True),
    )
    job_types = sort_jobs(job_types, options)
    columnWidths = compute_column_widths(job_types, worker, job, options)
    out_str += "Worker\n"
    out_str += sprint_tag("worker", worker, options, columnWidths=columnWidths)
    out_str += "Job\n"
    out_str += sprint_tag("job", job, options, columnWidths=columnWidths)
    for t in job_types:
        out_str += f" {t.name}\n"
        out_str += f"    Total Cores: {t.total_cores}\n"
        out_str += sprint_tag(t.name, t, options, columnWidths=columnWidths)
    return out_str


def compute_column_widths(
    job_types: List[Any], worker: Expando, job: Expando, options: Namespace
) -> ColumnWidths:
    """Return a ColumnWidths() object with the correct max widths."""
    cw = ColumnWidths()
    for t in job_types:
        update_column_widths(t, cw, options)
    update_column_widths(worker, cw, options)
    update_column_widths(job, cw, options)
    return cw


def update_column_widths(tag: Expando, cw: ColumnWidths, options: Namespace) -> None:
    """Update the column width attributes for this tag's fields."""
    # TODO: Deduplicate with actual printing code!
    for category in CATEGORIES:
        if category in options.categories:
            for field in ["min", "med", "ave", "max", "total"]:
                t = getattr(tag, f"{LONG_FORMS[field]}_{category}")
                width = cw.get_width(category, field)
                s = report(t, category, options, field=width).strip()
                if len(s) >= cw.get_width(category, field):
                    # this string is larger than max, width must be increased
                    cw.set_width(category, field, len(s) + 1)


def build_element(element: Expando, items: List[Job], item_name: str, defaults: dict[str, float]) -> Expando:
    """Create an element for output."""

    def assertNonnegative(i: float, name: str) -> float:
        if i < 0:
            raise RuntimeError("Negative value %s reported for %s" % (i, name))
        else:
            return float(i)

    # Make lists of all values for all items in each category, plus requested cores.
    item_values = {category: [] for category in (CATEGORIES + ["cores"])}

    for item in items:
        # If something lacks an entry, assume it used none of that thing.
        # This avoids crashing when jobs e.g. aren't done.
        for category, values in item_values.items():
            if category in COMPUTED_CATEGORIES:
                continue
            category_key = category if category != "cores" else "requested_cores"
            category_value = assertNonnegative(float(item.get(category_key, defaults[category])), category)
            values.append(category_value)

    for index in range(0, len(item_values[CATEGORIES[0]])):
        # For each item, compute the computed categories
        item_values["wait"].append(item_values["time"][index] * item_values["cores"][index] - item_values["clock"][index])

    for category, values in item_values.items():
        values.sort()

    if len(item_values[CATEGORIES[0]]) == 0:
        # Nothing actually there so make a 0 value
        for k, v in item_values.items():
            v.append(0)

    item_element = Expando(
        total_number=float(len(items)),
        name=item_name
    )

    for category, values in item_values.items():
        item_element["total_" + category] = float(sum(values))
        item_element["median_" + category] = float(values[len(values) // 2])
        item_element["average_" + category] = float(sum(values) / len(values))
        item_element["min_" + category] = float(min(values))
        item_element["max_" + category] = float(max(values))

    element[item_name] = item_element

    return item_element


def create_summary(
    element: Expando,
    containingItems: List[Expando],
    containingItemName: str,
    count_contained: Callable[[Expando], int],
) -> None:
    """
    Figure out how many jobs (or contained items) ran on each worker (or containing item).

    Stick a bunch of xxx_number_per_xxx stats into element to describe this.

    :param count_contained: function that maps from containing item to number of contained items.
    """

    # TODO: this still thinks like the old XML stats, even though now the
    # worker records no longer actually contain the job records.

    itemCounts = [count_contained(containingItem) for containingItem in containingItems]
    itemCounts.sort()
    if len(itemCounts) == 0:
        itemCounts.append(0)
    element["median_number_per_%s" % containingItemName] = itemCounts[
        len(itemCounts) // 2
    ]
    element["average_number_per_%s" % containingItemName] = float(
        sum(itemCounts) / len(itemCounts)
    )
    element["min_number_per_%s" % containingItemName] = min(itemCounts)
    element["max_number_per_%s" % containingItemName] = max(itemCounts)


def get_stats(jobStore: AbstractJobStore) -> Expando:
    """
    Sum together all the stats information in the job store.

    Produces  one object containing lists of the values from all the summed objects.
    """

    def aggregate_stats(fileHandle: TextIO, aggregateObject: Expando) -> None:
        try:
            stats = json.load(fileHandle, object_hook=Expando)
            for key in list(stats.keys()):
                if key in aggregateObject:
                    aggregateObject[key].append(stats[key])
                else:
                    aggregateObject[key] = [stats[key]]
        except ValueError:
            logger.critical(
                "File %s contains corrupted json. Skipping file." % fileHandle
            )
            pass  # The file is corrupted.

    aggregateObject = Expando()
    callBack = partial(aggregate_stats, aggregateObject=aggregateObject)
    jobStore.read_logs(callBack, read_all=True)
    return aggregateObject


def process_data(config: Config, stats: Expando) -> Expando:
    """
    Collate the stats and report
    """
    if "total_time" not in stats or "total_clock" not in stats:
        # toil job not finished yet
        stats.total_time = [0.0]
        stats.total_clock = [0.0]

    # This is actually the sum of *overall* wall clock time as measured by the
    # leader in each leader invocation, not a sum over jobs.
    stats.total_time = sum(float(number) for number in stats.total_time)
    # And this is CPU clock as measured by the leader, so it will count time
    # used in local jobs but not remote ones.
    stats.total_clock = sum(float(number) for number in stats.total_clock)

    collatedStatsTag = Expando(
        total_run_time=stats.total_time,
        total_clock=stats.total_clock,
        batch_system=config.batchSystem,
        default_memory=str(config.defaultMemory),
        default_cores=str(config.defaultCores),
        max_cores=str(config.maxCores if config.maxCores != SYS_MAX_SIZE else None),
    )

    # Add worker info
    worker = [_f for _f in getattr(stats, "workers", []) if _f]
    jobs = [_f for _f in getattr(stats, "jobs", []) if _f]
    jobs = [item for sublist in jobs for item in sublist]

    # Work out what usage to assume for things that didn't report
    defaults = {category: 0 for category in CATEGORIES}
    defaults["cores"] = config.defaultCores

    build_element(collatedStatsTag, worker, "worker", defaults)
    create_summary(
        build_element(collatedStatsTag, jobs, "jobs", defaults),
        getattr(stats, "workers", []),
        "worker",
        lambda worker: getattr(worker, "jobs_run", 0)
    )
    # Get info for each job
    jobNames = set()
    for job in jobs:
        jobNames.add(job.class_name)
    jobTypesTag = Expando()
    collatedStatsTag.job_types = jobTypesTag
    for jobName in jobNames:
        jobTypes = [job for job in jobs if job.class_name == jobName]
        build_element(jobTypesTag, jobTypes, jobName, defaults)
    collatedStatsTag.name = "collatedStatsTag"
    return collatedStatsTag


def report_data(tree: Expando, options: Namespace) -> None:
    # Now dump it all out to file
    if options.raw:
        out_str = json.dumps(tree, indent=4, separators=(",", ": "))
    else:
        out_str = report_pretty_data(
            tree, tree.worker, tree.jobs, tree.job_types.values(), options
        )
    if options.outputFile is not None:
        with open(options.outputFile, "w") as f:
            f.write(out_str)
    # Now dump onto the screen
    print(out_str)


sort_category_choices = CATEGORIES + ["alpha", "count"]
sort_field_choices = ["min", "med", "ave", "max", "total"]


def add_stats_options(parser: ArgumentParser) -> None:
    parser.add_argument(
        "--outputFile",
        dest="outputFile",
        default=None,
        help="File in which to write results.",
    )
    parser.add_argument(
        "--raw", action="store_true", default=False, help="Return raw json data."
    )
    parser.add_argument(
        "--pretty",
        "--human",
        action="store_true",
        default=False,
        help="if not raw, prettify the numbers to be human readable.",
    )
    parser.add_argument(
        "--sort",
        default="decending",
        choices=["ascending", "decending"],
        help="Sort direction.",
    )
    parser.add_argument(
        "--categories",
        default=",".join(CATEGORIES),
        type=str,
        help=f"Comma separated list of any of the following: {CATEGORIES}.",
    )
    parser.add_argument(
        "--sortCategory",
        default="time",
        choices=sort_category_choices,
        help=f"How to sort job categories.",
    )
    parser.add_argument(
        "--sortField",
        default="med",
        choices=sort_field_choices,
        help=f"How to sort job fields.",
    )


def main() -> None:
    """Reports stats on the workflow, use with --stats option to toil."""
    parser = parser_with_common_options(prog="toil stats")
    add_stats_options(parser)
    options = parser.parse_args()

    for c in options.categories.split(","):
        if c.strip().lower() not in CATEGORIES:
            logger.critical("Cannot use category %s, options are: %s", c.strip().lower(), CATEGORIES)
            sys.exit(1)
    options.categories = [x.strip().lower() for x in options.categories.split(",")]

    set_logging_from_options(options)
    config = Config()
    config.setOptions(options)
    try:
        jobStore = Toil.resumeJobStore(config.jobStore)
    except NoSuchJobStoreException:
        logger.critical("The job store %s does not exist", config.jobStore)
        sys.exit(1)
    stats = get_stats(jobStore)
    collatedStatsTag = process_data(jobStore.config, stats)
    report_data(collatedStatsTag, options)



# Copyright (C) 2015-2018 Regents of the University of California
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# 5.14.2018: copied into Toil from https://github.com/BD2KGenomics/bd2k-python-lib

import sys


# TODO: isn't this built in to Python 3 now?
class panic:
    """
    The Python idiom for reraising a primary exception fails when the except block raises a
    secondary exception, e.g. while trying to cleanup. In that case the original exception is
    lost and the secondary exception is reraised. The solution seems to be to save the primary
    exception info as returned from sys.exc_info() and then reraise that.

    This is a contextmanager that should be used like this

    try:
         # do something that can fail
    except:
        with panic( log ):
            # do cleanup that can also fail

    If a logging logger is passed to panic(), any secondary Exception raised within the with
    block will be logged. Otherwise those exceptions are swallowed. At the end of the with block
    the primary exception will be reraised.
    """

    def __init__( self, log=None ):
        super().__init__( )
        self.log = log
        self.exc_info = None

    def __enter__( self ):
        self.exc_info = sys.exc_info( )

    def __exit__( self, *exc_info ):
        if self.log is not None and exc_info and exc_info[ 0 ]:
            self.log.warning( "Exception during panic", exc_info=exc_info )
        exc_type, exc_value, traceback = self.exc_info
        raise_(exc_type, exc_value, traceback)

def raise_(exc_type, exc_value, traceback) -> None:
    if exc_value is not None:
        exc = exc_value
    else:
        exc = exc_type
    if exc.__traceback__ is not traceback:
        raise exc.with_traceback(traceback)
    raise exc



# Copyright (C) 2015-2021 Regents of the University of California
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# 5.14.2018: copied into Toil from https://github.com/BD2KGenomics/bd2k-python-lib
import datetime
import re
from functools import lru_cache, wraps
from threading import Lock
from typing import Any, Callable, Dict, Tuple, TypeVar

memoize = lru_cache(maxsize=None)
"""
Memoize a function result based on its parameters using this decorator.

For example, this can be used in place of lazy initialization. If the decorating
function is invoked by multiple threads, the decorated function may be called
more than once with the same arguments.
"""

MAT = TypeVar("MAT")
MRT = TypeVar("MRT")

def sync_memoize(f: Callable[[MAT], MRT]) -> Callable[[MAT], MRT]:
    """
    Like memoize, but guarantees that decorated function is only called once, even when multiple
    threads are calling the decorating function with multiple parameters.
    """
    # TODO: Think about an f that is recursive
    memory: Dict[Tuple[Any, ...], Any] = {}
    lock = Lock()

    @wraps(f)
    def new_f(*args: Any) -> Any:
        try:
            return memory[args]
        except KeyError:
            # on cache misses, retry with lock held
            with lock:
                try:
                    return memory[args]
                except KeyError:
                    r = f(*args)
                    memory[args] = r
                    return r
    return new_f


def parse_iso_utc(s: str) -> datetime.datetime:
    """
    Parses an ISO time with a hard-coded Z for zulu-time (UTC) at the end. Other timezones are
    not supported. Returns a timezone-naive datetime object. 

    :param s: The ISO-formatted time

    :return: A timezone-naive datetime object

    >>> parse_iso_utc('2016-04-27T00:28:04.000Z')
    datetime.datetime(2016, 4, 27, 0, 28, 4)
    >>> parse_iso_utc('2016-04-27T00:28:04Z')
    datetime.datetime(2016, 4, 27, 0, 28, 4)
    >>> parse_iso_utc('2016-04-27T00:28:04X')
    Traceback (most recent call last):
    ...
    ValueError: Not a valid ISO datetime in UTC: 2016-04-27T00:28:04X
    """
    rfc3339_datetime = re.compile(r'^(\d{4})-(\d{2})-(\d{2})T(\d{2}):(\d{2}):(\d{2})(?:\.(\d+))?(Z|[+-]\d{2}:\d{2})$')
    m = rfc3339_datetime.match(s)
    if not m:
        raise ValueError(f'Not a valid ISO datetime in UTC: {s}')
    else:
        fmt = '%Y-%m-%dT%H:%M:%S' + ('.%f' if m.group(7) else '') + 'Z'
        return datetime.datetime.strptime(s, fmt)


def strict_bool(s: str) -> bool:
    """Variant of bool() that only accepts two possible string values."""
    if s == 'True':
        return True
    elif s == 'False':
        return False
    else:
        raise ValueError(s)



# Copyright (C) 2015-2021 Regents of the University of California
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# 5.14.2018: copied into Toil from https://github.com/BD2KGenomics/bd2k-python-lib
from typing import Any, Iterable, Iterator, TypeVar

IT = TypeVar("IT")


def flatten(iterables: Iterable[IT]) -> Iterator[IT]:
    """Flatten an iterable, except for string elements."""
    for it in iterables:
        if isinstance(it, str):
            yield it
        else:
            yield from it


# noinspection PyPep8Naming
class concat:
    """
    A literal iterable to combine sequence literals (lists, set) with generators or list comprehensions.

    Instead of

    >>> [ -1 ] + [ x * 2 for x in range( 3 ) ] + [ -1 ]
    [-1, 0, 2, 4, -1]

    you can write

    >>> list( concat( -1, ( x * 2 for x in range( 3 ) ), -1 ) )
    [-1, 0, 2, 4, -1]

    This is slightly shorter (not counting the list constructor) and does not involve array
    construction or concatenation.

    Note that concat() flattens (or chains) all iterable arguments into a single
    result iterable:

    >>> list( concat( 1, range( 2, 4 ), 4 ) )
    [1, 2, 3, 4]

    It only does so one level deep. If you need to recursively flatten a data structure,
    check out crush().

    If you want to prevent that flattening for an iterable argument, wrap it in concat():

    >>> list( concat( 1, concat( range( 2, 4 ) ), 4 ) )
    [1, range(2, 4), 4]

    Some more example.

    >>> list( concat() ) # empty concat
    []
    >>> list( concat( 1 ) ) # non-iterable
    [1]
    >>> list( concat( concat() ) ) # empty iterable
    []
    >>> list( concat( concat( 1 ) ) ) # singleton iterable
    [1]
    >>> list( concat( 1, concat( 2 ), 3 ) ) # flattened iterable
    [1, 2, 3]
    >>> list( concat( 1, [2], 3 ) ) # flattened iterable
    [1, 2, 3]
    >>> list( concat( 1, concat( [2] ), 3 ) ) # protecting an iterable from being flattened
    [1, [2], 3]
    >>> list( concat( 1, concat( [2], 3 ), 4 ) ) # protection only works with a single argument
    [1, 2, 3, 4]
    >>> list( concat( 1, 2, concat( 3, 4 ), 5, 6 ) )
    [1, 2, 3, 4, 5, 6]
    >>> list( concat( 1, 2, concat( [ 3, 4 ] ), 5, 6 ) )
    [1, 2, [3, 4], 5, 6]

    Note that while strings are technically iterable, concat() does not flatten them.

    >>> list( concat( 'ab' ) )
    ['ab']
    >>> list( concat( concat( 'ab' ) ) )
    ['ab']
    """

    def __init__(self, *args: Any) -> None:
        super().__init__()
        self.args = args

    def __iter__(self) -> Iterator[Any]:
        def expand(x):
            if isinstance(x, concat) and len(x.args) == 1:
                i = x.args
            elif not isinstance(x, str):
                try:
                    i = x.__iter__()
                except AttributeError:
                    i = x,
            else:
                i = x
            return i

        return flatten(map(expand, self.args))



# Copyright (C) 2015-2021 Regents of the University of California
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# 5.14.2018: copied into Toil from https://github.com/BD2KGenomics/bd2k-python-lib
"""
This file holds the retry() decorator function and RetryCondition object.

retry() can be used to decorate any function based on the list of errors one
wishes to retry on.

This list of errors can contain normal Exception objects, and/or RetryCondition
objects wrapping Exceptions to include additional conditions.

For example, retrying on a one Exception (HTTPError)::

    from requests import get
    from requests.exceptions import HTTPError

    @retry(errors=[HTTPError])
    def update_my_wallpaper():
        return get('https://www.deviantart.com/')

Or::

    from requests import get
    from requests.exceptions import HTTPError

    @retry(errors=[HTTPError, ValueError])
    def update_my_wallpaper():
        return get('https://www.deviantart.com/')

The examples above will retry for the default interval on any errors specified
the "errors=" arg list.

To retry on specifically 500/502/503/504 errors, you could specify an ErrorCondition
object instead, for example::

    from requests import get
    from requests.exceptions import HTTPError

    @retry(errors=[
        ErrorCondition(
                   error=HTTPError,
                   error_codes=[500, 502, 503, 504]
               )])
    def update_my_wallpaper():
        return requests.get('https://www.deviantart.com/')

To retry on specifically errors containing the phrase "NotFound"::

    from requests import get
    from requests.exceptions import HTTPError

    @retry(errors=[
        ErrorCondition(
            error=HTTPError,
            error_message_must_include="NotFound"
        )])
    def update_my_wallpaper():
        return requests.get('https://www.deviantart.com/')

To retry on all HTTPError errors EXCEPT an HTTPError containing the phrase "NotFound"::

    from requests import get
    from requests.exceptions import HTTPError

    @retry(errors=[
        HTTPError,
        ErrorCondition(
                   error=HTTPError,
                   error_message_must_include="NotFound",
                   retry_on_this_condition=False
               )])
    def update_my_wallpaper():
        return requests.get('https://www.deviantart.com/')

To retry on boto3's specific status errors, an example of the implementation is::

    import boto3
    from botocore.exceptions import ClientError

    @retry(errors=[
        ErrorCondition(
                   error=ClientError,
                   boto_error_codes=["BucketNotFound"]
               )])
    def boto_bucket(bucket_name):
        boto_session = boto3.session.Session()
        s3_resource = boto_session.resource('s3')
        return s3_resource.Bucket(bucket_name)

Any combination of these will also work, provided the codes are matched to the correct
exceptions.  A ValueError will not return a 404, for example.

The retry function as a decorator should make retrying functions easier and clearer
It also encourages smaller independent functions, as opposed to lumping many different
things that may need to be retried on different conditions in the same function.

The ErrorCondition object tries to take some of the heavy lifting of writing specific
retry conditions and boil it down to an API that covers all common use-cases without
the user having to write any new bespoke functions.

Use-cases covered currently:

1. Retrying on a normal error, like a KeyError.
2. Retrying on HTTP error codes (use ErrorCondition).
3. Retrying on boto 3's specific status errors,
   like "BucketNotFound" (use ErrorCondition).
4. Retrying when an error message contains a certain phrase (use ErrorCondition).
5. Explicitly NOT retrying on a condition (use ErrorCondition).

If new functionality is needed, it's currently best practice in Toil to add
functionality to the ErrorCondition itself rather than making a new custom retry method.
"""
import copy
import functools
import http.client
import logging
import sqlite3
import time
import traceback
import urllib.error
from contextlib import contextmanager
from typing import (Any,
                    Callable,
                    ContextManager,
                    Generator,
                    Iterable,
                    List,
                    Optional,
                    Sequence,
                    Tuple,
                    Type,
                    Union)

import requests.exceptions
import urllib3.exceptions

SUPPORTED_HTTP_ERRORS = [http.client.HTTPException,
                         urllib.error.HTTPError,
                         urllib3.exceptions.HTTPError,
                         requests.exceptions.HTTPError]

try:
    import kubernetes.client.rest
    SUPPORTED_HTTP_ERRORS.append(kubernetes.client.rest.ApiException)
except ModuleNotFoundError:
    kubernetes = None

try:
    import botocore.exceptions
    SUPPORTED_HTTP_ERRORS.append(botocore.exceptions.ClientError)
except ModuleNotFoundError:
    botocore = None

logger = logging.getLogger(__name__)


class ErrorCondition:
    """
    A wrapper describing an error condition.

    ErrorCondition events may be used to define errors in more detail to determine
    whether to retry.
    """

    def __init__(self,
                 error: Optional[Any] = None,
                 error_codes: List[int] = None,
                 boto_error_codes: List[str] = None,
                 error_message_must_include: str = None,
                 retry_on_this_condition: bool = True):
        """
        Initialize this ErrorCondition.

        :param error: An Exception class to restrict to.
                      Must be specified if error_codes is used.
        :param error_codes: Numeric error codes (e.g. 404, 500, etc.) that must match
                            to be retried (optional; defaults to not checking).
        :param boto_error_codes: Human-readable error codes
                                 (e.g. "BucketNotFound", "ClientError", etc.) that are
                                 specific to Boto 3 and must match to be retried
                                 (optional; defaults to not checking).
        :param error_message_must_include: A string that must be in the error message
               to be retried (optional; defaults to not checking)
        :param retry_on_this_condition: This can be set to False to always error on
                                        this condition.
        """
        if error is None:
            if boto_error_codes:
                # Default to the base Boto 3 error
                error = botocore.exceptions.ClientError
            else:
                # Default to base Exception
                error = Exception

        self.error = error
        self.error_codes = error_codes
        self.boto_error_codes = boto_error_codes
        self.error_message_must_include = error_message_must_include
        self.retry_on_this_condition = retry_on_this_condition

        if self.error_codes:
            if error not in SUPPORTED_HTTP_ERRORS:
                raise NotImplementedError(
                    f"Unknown error type used with error_codes: {error}"
                )

        if self.boto_error_codes:
            if not issubclass(error, botocore.exceptions.ClientError):
                raise NotImplementedError(
                    f"Unknown error type used with boto_error_codes: {error}"
                )


def retry(
    intervals: Optional[List] = None,
    infinite_retries: bool = False,
    errors: Optional[Sequence[Union[ErrorCondition, Type[Exception]]]] = None,
    log_message: Optional[Tuple[Callable, str]] = None,
    prepare: Optional[List[Callable]] = None,
) -> Callable[[Any], Any]:
    """
    Retry a function if it fails with any Exception defined in "errors".

    Does so every x seconds, where x is defined by a list of numbers (ints or
    floats) in "intervals".  Also accepts ErrorCondition events
    for more detailed retry attempts.

    :param intervals: A list of times in seconds we keep retrying until returning failure.
        Defaults to retrying with the following exponential back-off before failing:
        1s, 1s, 2s, 4s, 8s, 16s
    :param infinite_retries: If this is True, reset the intervals when they run out.
        Defaults to: False.
    :param errors: A list of exceptions OR ErrorCondition objects to catch and retry on.
        ErrorCondition objects describe more detailed error event conditions than a plain error.
        An ErrorCondition specifies:
        - Exception (required)
        - Error codes that must match to be retried (optional; defaults to not checking)
        - A string that must be in the error message to be retried (optional; defaults to not checking)
        - A bool that can be set to False to always error on this condition.

        If not specified, this will default to a generic Exception.
    :param log_message: Optional tuple of ("log/print function()", "message string")
        that will precede each attempt.
    :param prepare: Optional list of functions to call, with the function's
        arguments, between retries, to reset state.
    :return: The result of the wrapped function or raise.
    """
    # set mutable defaults
    intervals = intervals if intervals else [1, 1, 2, 4, 8, 16]
    errors = errors if errors else [Exception]

    error_conditions = {error for error in errors if isinstance(error, ErrorCondition)}
    retriable_errors = {error for error in errors if not isinstance(error, ErrorCondition)}

    if log_message:
        post_message_function = log_message[0]
        message = log_message[1]

    # if a generic error exists (with no restrictions),
    # delete more specific error_condition instances of it
    for error_condition in error_conditions:
        if error_condition.retry_on_this_condition and error_condition.error in retriable_errors:
            error_conditions.remove(error_condition)

    # if a more specific error exists that isn't in the general set,
    # add it to the total errors that will be try/except-ed upon
    for error_condition in error_conditions:
        if error_condition.retry_on_this_condition:
            retriable_errors.add(error_condition.error)

    def decorate(func):
        @functools.wraps(func)
        def call(*args, **kwargs):
            intervals_remaining = copy.deepcopy(intervals)
            while True:
                try:
                    if log_message:
                        post_message_function(message)
                    return func(*args, **kwargs)

                except tuple(retriable_errors) as e:
                    if not intervals_remaining:
                        if infinite_retries:
                            intervals_remaining = copy.deepcopy(intervals)
                        else:
                            raise

                    if error_conditions:
                        if not error_meets_conditions(e, error_conditions):
                            raise

                    interval = intervals_remaining.pop(0)
                    logger.warning(f"Error in {func}: {e}. Retrying after {interval} s...")
                    time.sleep(interval)
                    if prepare is not None:
                        for prep_function in prepare:
                            # Reset state for next attempt
                            prep_function(*args, **kwargs)
        return call
    return decorate


def return_status_code(e):
    if kubernetes:
        if isinstance(e, kubernetes.client.rest.ApiException):
            return e.status

    if botocore:
        if isinstance(e, botocore.exceptions.ClientError):
            return e.response.get('ResponseMetadata', {}).get('HTTPStatusCode')

    if isinstance(e, requests.exceptions.HTTPError):
        return e.response.status_code
    elif isinstance(e, http.client.HTTPException) or \
            isinstance(e, urllib3.exceptions.HTTPError):
        return e.status
    elif isinstance(e, urllib.error.HTTPError):
        return e.code
    else:
        raise ValueError(f'Unsupported error type; cannot grok status code: {e}.')


def get_error_code(e: Exception) -> str:
    """
    Get the error code name from a Boto 2 or 3 error, or compatible types.

    Returns empty string for other errors.
    """
    if hasattr(e, 'error_code') and isinstance(e.error_code, str):
        # A Boto 2 error
        return e.error_code
    if hasattr(e, 'code') and isinstance(e.code, str):
        # A (different?) Boto 2 error
        return e.code
    elif hasattr(e, 'response') and hasattr(e.response, 'get'):
        # A Boto 3 error
        code = e.response.get('Error', {}).get('Code')
        if isinstance(code, str):
            return code
        else:
            return ''
    else:
        return ''


def get_error_message(e: Exception) -> str:
    """
    Get the error message string from a Boto 2 or 3 error, or compatible types.

    Note that error message conditions also check more than this; this function
    does not fall back to the traceback for incompatible types.
    """
    if hasattr(e, 'error_message') and isinstance(e.error_message, str):
        # A Boto 2 error
        return e.error_message
    elif hasattr(e, 'response') and hasattr(e.response, 'get'):
        # A Boto 3 error
        message = e.response.get('Error', {}).get('Message')
        if isinstance(message, str):
            return message
        else:
            return ''
    else:
        return ''


def get_error_status(e: Exception) -> int:
    """
    Get the HTTP status code from a compatible source.

    Such as a Boto 2 or 3 error,
    kubernetes.client.rest.ApiException, http.client.HTTPException,
    urllib3.exceptions.HTTPError, requests.exceptions.HTTPError,
    urllib.error.HTTPError, or compatible type

    Returns 0 from other errors.
    """
    def numify(x):
        """Make sure a value is an integer."""
        return int(str(x).strip())

    if hasattr(e, 'status'):
        # A Boto 2 error, kubernetes.client.rest.ApiException,
        # http.client.HTTPException, or urllib3.exceptions.HTTPError
        return numify(e.status)
    elif hasattr(e, 'response'):
        if hasattr(e.response, 'status_code'):
            # A requests.exceptions.HTTPError
            return numify(e.response.status_code)
        elif hasattr(e.response, 'get'):
            # A Boto 3 error
            return numify(e.response.get('ResponseMetadata', {}).get('HTTPStatusCode'))
    elif hasattr(e, 'code'):
        # A urllib.error.HTTPError
        return numify(e.code)
    else:
        return 0


def get_error_body(e: Exception) -> str:
    """
    Get the body from a Boto 2 or 3 error, or compatible types.

    Returns the code and message if the error does not have a body.
    """
    if hasattr(e, 'body'):
        # A Boto 2 error
        if isinstance(e.body, bytes):
            # Decode the body first
            return e.body.decode('utf-8')
        elif isinstance(e.body, str):
            return e.body

    # Anything else
    return f'{get_error_code(e)}: {get_error_message(e)}'

def meets_error_message_condition(e: Exception, error_message: Optional[str]):
    if error_message:
        if kubernetes:
            if isinstance(e, kubernetes.client.rest.ApiException):
                return error_message in str(e)

        if botocore:
            if isinstance(e, botocore.exceptions.ClientError):
                return error_message in str(e)

        if isinstance(e, http.client.HTTPException) or isinstance(e, urllib3.exceptions.HTTPError):
            return error_message in e.reason
        elif isinstance(e, sqlite3.OperationalError):
            return error_message in str(e)
        elif isinstance(e, urllib.error.HTTPError):
            return error_message in e.msg
        elif isinstance(e, requests.exceptions.HTTPError):
            return error_message in e.raw
        elif hasattr(e, 'msg'):
            return error_message in e.msg
        else:
            return error_message in traceback.format_exc()
    else:  # there is no error message, so the user is not expecting it to be present
        return True


def meets_error_code_condition(e: Exception, error_codes: Optional[List[int]]):
    """These are expected to be normal HTTP error codes, like 404 or 500."""
    if error_codes:
        status_code = get_error_status(e)
        return status_code in error_codes
    else:  # there are no error codes, so the user is not expecting the status to match
        return True


def meets_boto_error_code_condition(e: Exception, boto_error_codes: Optional[List[str]]):
    """These are expected to be AWS's custom error aliases, like 'BucketNotFound' or 'AccessDenied'."""
    if boto_error_codes:
        status_code = get_error_code(e)
        return status_code in boto_error_codes
    else:  # there are no boto error codes, so the user is not expecting the status to match
        return True


def error_meets_conditions(e, error_conditions):
    condition_met = False
    for error in error_conditions:
        if isinstance(e, error.error):
            if error.error_codes or error.boto_error_codes or error.error_message_must_include:
                error_message_condition_met = meets_error_message_condition(e, error.error_message_must_include)
                error_code_condition_met = meets_error_code_condition(e, error.error_codes)
                boto_error_code_condition_met = meets_boto_error_code_condition(e, error.boto_error_codes)
                if error_message_condition_met and error_code_condition_met and boto_error_code_condition_met:
                    if not error.retry_on_this_condition:
                        return False
                    condition_met = True
    return condition_met

DEFAULT_DELAYS = (0, 1, 1, 4, 16, 64)
DEFAULT_TIMEOUT = 300

# TODO: Replace the use of this with retry()
#  The aws provisioner and jobstore need a large refactoring to be boto3 compliant, so this is
#  still used there to avoid the duplication of future work
def old_retry(
    delays: Iterable[float] = DEFAULT_DELAYS,
    timeout: float = DEFAULT_TIMEOUT,
    predicate: Callable[[Exception], bool] = lambda e: False,
) -> Generator[ContextManager, None, None]:
    """
    Deprecated.

    Retry an operation while the failure matches a given predicate and until a given timeout
    expires, waiting a given amount of time in between attempts. This function is a generator
    that yields contextmanagers. See doctests below for example usage.

    :param Iterable[float] delays: an interable yielding the time in seconds to wait before each
           retried attempt, the last element of the iterable will be repeated.

    :param float timeout: a overall timeout that should not be exceeded for all attempts together.
           This is a best-effort mechanism only and it won't abort an ongoing attempt, even if the
           timeout expires during that attempt.

    :param Callable[[Exception],bool] predicate: a unary callable returning True if another
           attempt should be made to recover from the given exception. The default value for this
           parameter will prevent any retries!

    :return: a generator yielding context managers, one per attempt
    :rtype: Iterator

    Retry for a limited amount of time:

    >>> true = lambda _:True
    >>> false = lambda _:False
    >>> i = 0
    >>> for attempt in old_retry( delays=[0], timeout=.1, predicate=true ):
    ...     with attempt:
    ...         i += 1
    ...         raise RuntimeError('foo')
    Traceback (most recent call last):
    ...
    RuntimeError: foo
    >>> i > 1
    True

    If timeout is 0, do exactly one attempt:

    >>> i = 0
    >>> for attempt in old_retry( timeout=0 ):
    ...     with attempt:
    ...         i += 1
    ...         raise RuntimeError( 'foo' )
    Traceback (most recent call last):
    ...
    RuntimeError: foo
    >>> i
    1

    Don't retry on success:

    >>> i = 0
    >>> for attempt in old_retry( delays=[0], timeout=.1, predicate=true ):
    ...     with attempt:
    ...         i += 1
    >>> i
    1

    Don't retry on unless predicate returns True:

    >>> i = 0
    >>> for attempt in old_retry( delays=[0], timeout=.1, predicate=false):
    ...     with attempt:
    ...         i += 1
    ...         raise RuntimeError( 'foo' )
    Traceback (most recent call last):
    ...
    RuntimeError: foo
    >>> i
    1
    """
    if timeout > 0:
        go = [ None ]

        @contextmanager
        def repeated_attempt( delay ):
            try:
                yield
            except Exception as e:
                if time.time( ) + delay < expiration:
                    if predicate( e ):
                        logger.info('Got %s, trying again in %is.', e, delay)
                        time.sleep( delay )
                    else:
                        logger.error('Got a %s: %s which is not retriable according to %s', type(e), e, predicate)
                        raise
                else:
                    logger.error('Got %s and no time is left to retry', e)
                    raise
            else:
                go.pop( )

        delays = iter( delays )
        expiration = time.time( ) + timeout
        delay = next( delays )
        while go:
            yield repeated_attempt( delay )
            delay = next( delays, delay )
    else:
        @contextmanager
        def single_attempt( ):
            yield

        yield single_attempt( )

# Decorator to retry tests that fail. Needs to be called with
# prepare=[tearDown, setUp] if the test class has tear down and set up that
# needs to happen between tests.
retry_flaky_test = functools.partial(retry, intervals=[1, 1, 1])



# Copyright (C) 2015-2021 Regents of the University of California
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import errno
import logging
import os
import socket
import sys
from typing import (Any,
                    Callable,
                    ContextManager,
                    Dict,
                    Iterable,
                    Iterator,
                    List,
                    Optional,
                    Set,
                    Union,
                    cast)
from urllib.parse import ParseResult

from toil.lib.aws import session
from toil.lib.misc import printq
from toil.lib.retry import (DEFAULT_DELAYS,
                            DEFAULT_TIMEOUT,
                            get_error_code,
                            get_error_status,
                            old_retry,
                            retry)

if sys.version_info >= (3, 8):
    from typing import Literal
else:
    from typing_extensions import Literal

try:
    from boto.exception import BotoServerError, S3ResponseError
    from botocore.exceptions import ClientError
    from mypy_boto3_iam import IAMClient, IAMServiceResource
    from mypy_boto3_s3 import S3Client, S3ServiceResource
    from mypy_boto3_s3.literals import BucketLocationConstraintType
    from mypy_boto3_s3.service_resource import Bucket, Object
    from mypy_boto3_sdb import SimpleDBClient
except ImportError:
    BotoServerError = None  # type: ignore
    ClientError = None  # type: ignore
    # AWS/boto extra is not installed

logger = logging.getLogger(__name__)

# These are error codes we expect from AWS if we are making requests too fast.
# https://github.com/boto/botocore/blob/49f87350d54f55b687969ec8bf204df785975077/botocore/retries/standard.py#L316
THROTTLED_ERROR_CODES = [
        'Throttling',
        'ThrottlingException',
        'ThrottledException',
        'RequestThrottledException',
        'TooManyRequestsException',
        'ProvisionedThroughputExceededException',
        'TransactionInProgressException',
        'RequestLimitExceeded',
        'BandwidthLimitExceeded',
        'LimitExceededException',
        'RequestThrottled',
        'SlowDown',
        'PriorRequestNotComplete',
        'EC2ThrottledException',
]

@retry(errors=[BotoServerError])
def delete_iam_role(
    role_name: str, region: Optional[str] = None, quiet: bool = True
) -> None:
    from boto.iam.connection import IAMConnection

    # TODO: the Boto3 type hints are a bit oversealous here; they want hundreds
    # of overloads of the client-getting methods to exist based on the literal
    # string passed in, to return exactly the right kind of client or resource.
    # So we end up having to wrap all the calls in casts, which kind of defeats
    # the point of a nice fluent method you can call with the name of the thing
    # you want; we should have been calling iam_client() and so on all along if
    # we wanted MyPy to be able to understand us. So at some point we should
    # consider revising our API here to be less annoying to explain to the type
    # checker.
    iam_client = cast(IAMClient, session.client('iam', region_name=region))
    iam_resource = cast(IAMServiceResource, session.resource('iam', region_name=region))
    boto_iam_connection = IAMConnection()
    role = iam_resource.Role(role_name)
    # normal policies
    for attached_policy in role.attached_policies.all():
        printq(f'Now dissociating policy: {attached_policy.policy_name} from role {role.name}', quiet)
        role.detach_policy(PolicyArn=attached_policy.arn)
    # inline policies
    for inline_policy in role.policies.all():
        printq(f'Deleting inline policy: {inline_policy.policy_name} from role {role.name}', quiet)
        # couldn't find an easy way to remove inline policies with boto3; use boto
        boto_iam_connection.delete_role_policy(role.name, inline_policy.policy_name)
    iam_client.delete_role(RoleName=role_name)
    printq(f'Role {role_name} successfully deleted.', quiet)


@retry(errors=[BotoServerError])
def delete_iam_instance_profile(
    instance_profile_name: str, region: Optional[str] = None, quiet: bool = True
) -> None:
    iam_resource = cast(IAMServiceResource, session.resource("iam", region_name=region))
    instance_profile = iam_resource.InstanceProfile(instance_profile_name)
    if instance_profile.roles is not None:
        for role in instance_profile.roles:
            printq(f'Now dissociating role: {role.name} from instance profile {instance_profile_name}', quiet)
            instance_profile.remove_role(RoleName=role.name)
    instance_profile.delete()
    printq(f'Instance profile "{instance_profile_name}" successfully deleted.', quiet)


@retry(errors=[BotoServerError])
def delete_sdb_domain(
    sdb_domain_name: str, region: Optional[str] = None, quiet: bool = True
) -> None:
    sdb_client = cast(SimpleDBClient, session.client("sdb", region_name=region))
    sdb_client.delete_domain(DomainName=sdb_domain_name)
    printq(f'SBD Domain: "{sdb_domain_name}" successfully deleted.', quiet)


def connection_reset(e: Exception) -> bool:
    """
    Return true if an error is a connection reset error.
    """
    # For some reason we get 'error: [Errno 104] Connection reset by peer' where the
    # English description suggests that errno is 54 (ECONNRESET) while the actual
    # errno is listed as 104. To be safe, we check for both:
    return isinstance(e, socket.error) and e.errno in (errno.ECONNRESET, 104)

# TODO: Replace with: @retry and ErrorCondition
def retryable_s3_errors(e: Exception) -> bool:
    """
    Return true if this is an error from S3 that looks like we ought to retry our request.
    """
    return (connection_reset(e)
            or (isinstance(e, BotoServerError) and e.status in (429, 500))
            or (isinstance(e, BotoServerError) and e.code in THROTTLED_ERROR_CODES)
            # boto3 errors
            or (isinstance(e, (S3ResponseError, ClientError)) and get_error_code(e) in THROTTLED_ERROR_CODES)
            or (isinstance(e, ClientError) and 'BucketNotEmpty' in str(e))
            or (isinstance(e, ClientError) and e.response.get('ResponseMetadata', {}).get('HTTPStatusCode') == 409 and 'try again' in str(e))
            or (isinstance(e, ClientError) and e.response.get('ResponseMetadata', {}).get('HTTPStatusCode') in (404, 429, 500, 502, 503, 504)))


def retry_s3(delays: Iterable[float] = DEFAULT_DELAYS, timeout: float = DEFAULT_TIMEOUT, predicate: Callable[[Exception], bool] = retryable_s3_errors) -> Iterator[ContextManager[None]]:
    """
    Retry iterator of context managers specifically for S3 operations.
    """
    return old_retry(delays=delays, timeout=timeout, predicate=predicate)

@retry(errors=[BotoServerError])
def delete_s3_bucket(
    s3_resource: "S3ServiceResource",
    bucket: str,
    quiet: bool = True
) -> None:
    """
    Delete the given S3 bucket.
    """
    printq(f'Deleting s3 bucket: {bucket}', quiet)

    paginator = s3_resource.meta.client.get_paginator('list_object_versions')
    try:
        for response in paginator.paginate(Bucket=bucket):
            # Versions and delete markers can both go in here to be deleted.
            # They both have Key and VersionId, but there's no shared base type
            # defined for them in the stubs to express that. See
            # <https://github.com/vemel/mypy_boto3_builder/issues/123>. So we
            # have to do gymnastics to get them into the same list.
            to_delete: List[Dict[str, Any]] = cast(List[Dict[str, Any]], response.get('Versions', [])) + \
                                              cast(List[Dict[str, Any]], response.get('DeleteMarkers', []))
            for entry in to_delete:
                printq(f"    Deleting {entry['Key']} version {entry['VersionId']}", quiet)
                s3_resource.meta.client.delete_object(Bucket=bucket, Key=entry['Key'], VersionId=entry['VersionId'])
        s3_resource.Bucket(bucket).delete()
        printq(f'\n * Deleted s3 bucket successfully: {bucket}\n\n', quiet)
    except s3_resource.meta.client.exceptions.NoSuchBucket:
        printq(f'\n * S3 bucket no longer exists: {bucket}\n\n', quiet)


def create_s3_bucket(
    s3_resource: "S3ServiceResource",
    bucket_name: str,
    region: Union["BucketLocationConstraintType", Literal["us-east-1"]],
) -> "Bucket":
    """
    Create an AWS S3 bucket, using the given Boto3 S3 session, with the
    given name, in the given region.

    Supports the us-east-1 region, where bucket creation is special.

    *ALL* S3 bucket creation should use this function.
    """
    logger.debug("Creating bucket '%s' in region %s.", bucket_name, region)
    if region == "us-east-1":  # see https://github.com/boto/boto3/issues/125
        bucket = s3_resource.create_bucket(Bucket=bucket_name)
    else:
        bucket = s3_resource.create_bucket(
            Bucket=bucket_name,
            CreateBucketConfiguration={"LocationConstraint": region},
        )
    return bucket

@retry(errors=[ClientError])
def enable_public_objects(bucket_name: str) -> None:
    """
    Enable a bucket to contain objects which are public.

    This adjusts the bucket's Public Access Block setting to not block all
    public access, and also adjusts the bucket's Object Ownership setting to a
    setting which enables object ACLs.

    Does *not* touch the *account*'s Public Access Block setting, which can
    also interfere here. That is probably best left to the account
    administrator.

    This configuration used to be the default, and is what most of Toil's code
    is written to expect, but it was changed so that new buckets default to the
    more restrictive setting
    <https://aws.amazon.com/about-aws/whats-new/2022/12/amazon-s3-automatically-enable-block-public-access-disable-access-control-lists-buckets-april-2023/>,
    with the expectation that people would write IAM policies for the buckets
    to allow public access if needed. Toil expects to be able to make arbitrary
    objects in arbitrary places public, and naming them all in an IAM policy
    would be a very awkward way to do it. So we restore the old behavior.
    """

    s3_client = cast(S3Client, session.client('s3'))

    # Even though the new default is for public access to be prohibited, this
    # is implemented by adding new things attached to the bucket. If we remove
    # those things the bucket will default to the old defaults. See
    # <https://aws.amazon.com/blogs/aws/heads-up-amazon-s3-security-changes-are-coming-in-april-of-2023/>.

    # Stop blocking public access
    s3_client.delete_public_access_block(Bucket=bucket_name)

    # Stop using an ownership controls setting that prohibits ACLs.
    s3_client.delete_bucket_ownership_controls(Bucket=bucket_name)


def get_bucket_region(bucket_name: str, endpoint_url: Optional[str] = None, only_strategies: Optional[Set[int]] = None) -> str:
    """
    Get the AWS region name associated with the given S3 bucket.

    Takes an optional S3 API URL override.

    :param only_strategies: For testing, use only strategies with 1-based numbers in this set.
    """

    s3_client = cast(S3Client, session.client('s3', endpoint_url=endpoint_url))

    def attempt_get_bucket_location() -> Optional[str]:
        """
        Try and get the bucket location from the normal API call.
        """
        return s3_client.get_bucket_location(Bucket=bucket_name).get('LocationConstraint', None)

    def attempt_get_bucket_location_from_us_east_1() -> Optional[str]:
        """
        Try and get the bucket location from the normal API call, but against us-east-1
        """
        # Sometimes we aren't allowed to GetBucketLocation. At least some of
        # the time, that's only true when we talk to whatever S3 API servers we
        # usually use, and we can get around this lack of permission by talking
        # to us-east-1 instead. We've been told that this is because us-east-1
        # is special and will answer the question when other regions won't.
        # See:
        # <https://ucsc-gi.slack.com/archives/C027D41M6UA/p1652819831740169?thread_ts=1652817377.594539&cid=C027D41M6UA>
        # It could also be because AWS open data buckets (which we tend to
        # encounter this problem for) tend to actually themselves be in
        # us-east-1.
        backup_s3_client = cast(S3Client, session.client('s3', region_name='us-east-1'))
        return backup_s3_client.get_bucket_location(Bucket=bucket_name).get('LocationConstraint', None)

    def attempt_head_bucket() -> Optional[str]:
        """
        Try and get the bucket location from calling HeadBucket and inspecting
        the headers.
        """
        # If that also doesn't work, we can try HEAD-ing the bucket and looking
        # for an 'x-amz-bucket-region' header on the response, which can tell
        # us where the bucket is. See
        # <https://github.com/aws/aws-sdk-cpp/issues/844#issuecomment-383747871>
        info = s3_client.head_bucket(Bucket=bucket_name)
        return info['ResponseMetadata']['HTTPHeaders']['x-amz-bucket-region']

    # Compose a list of strategies we want to try in order, which may work.
    # None is an acceptable return type that actually means something.
    strategies: List[Callable[[], Optional[str]]] = []
    strategies.append(attempt_get_bucket_location)
    if not endpoint_url:
        # We should only try to talk to us-east-1 if we don't have a custom
        # URL.
        strategies.append(attempt_get_bucket_location_from_us_east_1)
    strategies.append(attempt_head_bucket)

    for attempt in retry_s3():
        with attempt:
            for i, strategy in enumerate(strategies):
                if only_strategies is not None and i+1 not in only_strategies:
                    # We want to test running without this strategy.
                    continue
                try:
                    return bucket_location_to_region(strategy())
                except ClientError as e:
                    if get_error_code(e) == 'AccessDenied' and not endpoint_url:
                        logger.warning('Strategy %d to get bucket location did not work: %s', i + 1, e)
                        last_error: Exception = e
                        # We were blocked with this strategy. Move on to the
                        # next strategy which might work.
                        continue
                    else:
                        raise
                except KeyError as e:
                    # If we get a weird head response we will have a KeyError
                    logger.warning('Strategy %d to get bucket location did not work: %s', i + 1, e)
                    last_error = e
    # If we get here we ran out of attempts. Raise whatever the last problem was.
    raise last_error

def region_to_bucket_location(region: str) -> str:
    return '' if region == 'us-east-1' else region

def bucket_location_to_region(location: Optional[str]) -> str:
    return "us-east-1" if location == "" or location is None else location

def get_object_for_url(url: ParseResult, existing: Optional[bool] = None) -> "Object":
        """
        Extracts a key (object) from a given parsed s3:// URL.

        If existing is true and the object does not exist, raises FileNotFoundError.

        :param bool existing: If True, key is expected to exist. If False, key is expected not to
                exists and it will be created. If None, the key will be created if it doesn't exist.
        """

        key_name = url.path[1:]
        bucket_name = url.netloc

        # Decide if we need to override Boto's built-in URL here.
        endpoint_url: Optional[str] = None
        host = os.environ.get('TOIL_S3_HOST', None)
        port = os.environ.get('TOIL_S3_PORT', None)
        protocol = 'https'
        if os.environ.get('TOIL_S3_USE_SSL', True) == 'False':
            protocol = 'http'
        if host:
            endpoint_url = f'{protocol}://{host}' + f':{port}' if port else ''

        # TODO: OrdinaryCallingFormat equivalent in boto3?
        # if botoargs:
        #     botoargs['calling_format'] = boto.s3.connection.OrdinaryCallingFormat()

        try:
            # Get the bucket's region to avoid a redirect per request
            region = get_bucket_region(bucket_name, endpoint_url=endpoint_url)
            s3 = cast(S3ServiceResource, session.resource('s3', region_name=region, endpoint_url=endpoint_url))
        except ClientError:
            # Probably don't have permission.
            # TODO: check if it is that
            s3 = cast(S3ServiceResource, session.resource('s3', endpoint_url=endpoint_url))

        obj = s3.Object(bucket_name, key_name)
        objExists = True

        try:
            obj.load()
        except ClientError as e:
            if get_error_status(e) == 404:
                objExists = False
            else:
                raise
        if existing is True and not objExists:
            raise FileNotFoundError(f"Key '{key_name}' does not exist in bucket '{bucket_name}'.")
        elif existing is False and objExists:
            raise RuntimeError(f"Key '{key_name}' exists in bucket '{bucket_name}'.")

        if not objExists:
            obj.put()  # write an empty file
        return obj


@retry(errors=[BotoServerError])
def list_objects_for_url(url: ParseResult) -> List[str]:
        """
        Extracts a key (object) from a given parsed s3:// URL. The URL will be
        supplemented with a trailing slash if it is missing.
        """
        key_name = url.path[1:]
        bucket_name = url.netloc

        if key_name != '' and not key_name.endswith('/'):
            # Make sure to put the trailing slash on the key, or else we'll see
            # a prefix of just it.
            key_name = key_name + '/'

        # Decide if we need to override Boto's built-in URL here.
        # TODO: Deduplicate with get_object_for_url, or push down into session module
        endpoint_url: Optional[str] = None
        host = os.environ.get('TOIL_S3_HOST', None)
        port = os.environ.get('TOIL_S3_PORT', None)
        protocol = 'https'
        if os.environ.get('TOIL_S3_USE_SSL', True) == 'False':
            protocol = 'http'
        if host:
            endpoint_url = f'{protocol}://{host}' + f':{port}' if port else ''

        client = cast(S3Client, session.client('s3', endpoint_url=endpoint_url))

        listing = []

        paginator = client.get_paginator('list_objects_v2')
        result = paginator.paginate(Bucket=bucket_name, Prefix=key_name, Delimiter='/')
        for page in result:
            if 'CommonPrefixes' in page:
                for prefix_item in page['CommonPrefixes']:
                    listing.append(prefix_item['Prefix'][len(key_name):])
            if 'Contents' in page:
                for content_item in page['Contents']:
                    if content_item['Key'] == key_name:
                        # Ignore folder name itself
                        continue
                    listing.append(content_item['Key'][len(key_name):])

        logger.debug('Found in %s items: %s', url, listing)
        return listing

def flatten_tags(tags: Dict[str, str]) -> List[Dict[str, str]]:
    """
    Convert tags from a key to value dict into a list of 'Key': xxx, 'Value': xxx dicts.
    """
    return [{'Key': k, 'Value': v} for k, v in tags.items()]



# Copyright (C) 2015-2021 Regents of the University of California
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import inspect
import logging
import os
import random
import sys
from uuid import uuid4

from toil.common import getNodeID
from toil.lib.exceptions import panic, raise_
from toil.lib.io import (AtomicFileCreate,
                         atomic_install,
                         atomic_tmp_file,
                         mkdtemp)
from toil.lib.misc import CalledProcessErrorStderr, call_command
from toil.test import ToilTest, slow

log = logging.getLogger(__name__)
logging.basicConfig()


class MiscTests(ToilTest):
    """
    This class contains miscellaneous tests that don't have enough content to be their own test
    file, and that don't logically fit in with any of the other test suites.
    """
    def setUp(self):
        super().setUp()
        self.testDir = self._createTempDir()

    def testIDStability(self):
        prevNodeID = None
        for i in range(10, 1):
            nodeID = getNodeID()
            self.assertEqual(nodeID, prevNodeID)
            prevNodeID = nodeID

    @slow
    def testGetSizeOfDirectoryWorks(self):
        '''A test to make sure toil.common.getDirSizeRecursively does not
        underestimate the amount of disk space needed.

        Disk space allocation varies from system to system.  The computed value
        should always be equal to or slightly greater than the creation value.
        This test generates a number of random directories and randomly sized
        files to test this using getDirSizeRecursively.
        '''
        from toil.common import getDirSizeRecursively

        # a list of the directories used in the test
        directories = [self.testDir]
        # A dict of {FILENAME: FILESIZE} for all files used in the test
        files = {}
        # Create a random directory structure
        for i in range(0,10):
            directories.append(mkdtemp(dir=random.choice(directories), prefix='test'))
        # Create 50 random file entries in different locations in the directories. 75% of the time
        # these are fresh files of size [1, 10] MB and 25% of the time they are hard links to old
        # files.
        while len(files) <= 50:
            fileName = os.path.join(random.choice(directories), self._getRandomName())
            if random.randint(0,100) < 75:
                # Create a fresh file in the range of 1-10 MB
                fileSize = int(round(random.random(), 2) * 10 * 1024 * 1024)
                with open(fileName, 'wb') as fileHandle:
                    fileHandle.write(os.urandom(fileSize))
                files[fileName] = fileSize
            else:
                # Link to one of the previous files
                if len(files) == 0:
                    continue
                linkSrc = random.choice(list(files.keys()))
                os.link(linkSrc, fileName)
                files[fileName] = 'Link to %s' % linkSrc

        computedDirectorySize = getDirSizeRecursively(self.testDir)
        totalExpectedSize = sum(x for x in list(files.values()) if isinstance(x, int))
        self.assertGreaterEqual(computedDirectorySize, totalExpectedSize)

    @staticmethod
    def _getRandomName():
        return uuid4().hex

    def _get_test_out_file(self, tail):
        outf = os.path.join(self.testDir, self.id() + "." + tail)
        if os.path.exists(outf):
            os.unlink(outf)
        return outf

    def _write_test_file(self, outf_tmp):
        with open(outf_tmp, "w") as fh:
            fh.write(self.id() + '\n')

    def test_atomic_install(self):
        outf = self._get_test_out_file(".foo.gz")
        outf_tmp = atomic_tmp_file(outf)
        self._write_test_file(outf_tmp)
        atomic_install(outf_tmp, outf)
        self.assertTrue(os.path.exists(outf))

    def test_atomic_install_dev(self):
        devn = '/dev/null'
        tmp = atomic_tmp_file(devn)
        self.assertEqual(tmp, devn)
        atomic_install(tmp, devn)

    def test_atomic_context_ok(self):
        outf = self._get_test_out_file(".tar")
        with AtomicFileCreate(outf) as outf_tmp:
            self._write_test_file(outf_tmp)
        self.assertTrue(os.path.exists(outf))

    def test_atomic_context_error(self):
        outf = self._get_test_out_file(".tar")
        try:
            with AtomicFileCreate(outf) as outf_tmp:
                self._write_test_file(outf_tmp)
                raise Exception("stop!")
        except Exception as ex:
            self.assertEqual(str(ex), "stop!")
        self.assertFalse(os.path.exists(outf))

    def test_call_command_ok(self):
        o = call_command(["echo", "Fred"])
        self.assertEqual("Fred\n", o)
        self.assertTrue(isinstance(o, str), str(type(o)))

    def test_call_command_err(self):
        with self.assertRaisesRegex(CalledProcessErrorStderr,
                                     "^Command '\\['cat', '/dev/Frankenheimer']' exit status 1: cat: /dev/Frankenheimer: No such file or directory\n$"):
            call_command(["cat", "/dev/Frankenheimer"])

class TestPanic(ToilTest):
    def test_panic_by_hand(self):
        try:
            self.try_and_panic_by_hand()
        except:
            self.__assert_raised_exception_is_primary()

    def test_panic(self):
        try:
            self.try_and_panic()
        except:
            self.__assert_raised_exception_is_primary()

    def test_panic_with_secondary(self):
        try:
            self.try_and_panic_with_secondary()
        except:
            self.__assert_raised_exception_is_primary()

    def test_nested_panic(self):
        try:
            self.try_and_nested_panic_with_secondary()
        except:
            self.__assert_raised_exception_is_primary()

    def try_and_panic_by_hand(self):
        try:
            self.line_of_primary_exc = inspect.currentframe().f_lineno + 1
            raise ValueError("primary")
        except Exception:
            exc_type, exc_value, traceback = sys.exc_info()
            try:
                raise RuntimeError("secondary")
            except Exception:
                pass
            raise_(exc_type, exc_value, traceback)

    def try_and_panic(self):
        try:
            self.line_of_primary_exc = inspect.currentframe().f_lineno + 1
            raise ValueError("primary")
        except:
            with panic(log):
                pass

    def try_and_panic_with_secondary(self):
        try:
            self.line_of_primary_exc = inspect.currentframe().f_lineno + 1
            raise ValueError("primary")
        except:
            with panic( log ):
                raise RuntimeError("secondary")

    def try_and_nested_panic_with_secondary(self):
        try:
            self.line_of_primary_exc = inspect.currentframe().f_lineno + 1
            raise ValueError("primary")
        except:
            with panic( log ):
                with panic( log ):
                    raise RuntimeError("secondary")

    def __assert_raised_exception_is_primary(self):
        exc_type, exc_value, exc_traceback = sys.exc_info()
        self.assertEqual(exc_type, ValueError)
        self.assertEqual(str(exc_value), "primary")
        while exc_traceback.tb_next is not None:
            exc_traceback = exc_traceback.tb_next
        self.assertEqual(exc_traceback.tb_lineno, self.line_of_primary_exc)
